import re
from uuid import UUID

from dcicutils.misc_utils import ignored
from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden
from pyramid.security import Authenticated
from pyramid.traversal import find_resource
from pyramid.view import view_config
from .util import debug_log

ATID_PATTERN = re.compile("/[a-zA-Z-]+/[a-zA-Z0-9-_:]+/")
GENELIST_ATID = re.compile("/gene-lists/[a-zA-Z0-9-]+/")  # TODO: refactor
MINIMAL_EMBEDS = ["projects", "institutions", "users"]  # TODO: refactor for setting in downstream repo
MINIMAL_EMBED_ATID = re.compile("/(" + "|".join(MINIMAL_EMBEDS) + ")/[a-zA-Z0-9-_:]+/")
KEYS_TO_IGNORE = [
    "@id",
    "@type",
    "principals_allowed",
    "uuid",
    "status",
    "title",
    "display_title",
    "schema_version",
    "date_created",
    "actions",
]
FORBIDDEN_MSG = {"error": "no view permissions"}
DATABASE_ITEM_KEY = "@type"  # Key specific to JSON objects that are CGAP items


def includeme(config):
    config.add_route("embed", "/embed")
    config.scan(__name__)


class CustomEmbed:
    """
    Class to handle custom embedding for /embed API.
    """

    def __init__(self, request, item, embed_props):
        self.request = request
        self.ignored_embeds = embed_props.get("ignored_embeds", [])
        self.desired_embeds = embed_props.get("desired_embeds", [])
        self.embed_depth = embed_props.get("embed_depth", 4)
        self.requested_fields = embed_props.get("requested_fields", [])

        self.cache = {}
        self.invalid_ids = []
        if self.requested_fields:
            self.nested_fields = self.fields_to_nested_dict()
            item = self.user_embed(item, initial_item=True)
            self.result = self.field_embed(item, self.nested_fields, initial_item=True)
        else:
            depth = -1
            self.result = self.embed(item, depth)

    def add_actions(self, item):
        """
        Add the "actions" field to an item according to the request's
        permissions, formatted identically to the calc props on items
        for a GET page view.

        :param item: dict item in object view
        :return item: dict item in object view with "actions"
        """
        actions = []
        root_resource = self.request.root
        item_path = item["@id"]
        item_resource = find_resource(root_resource, item_path)
        for action in ["edit", "create"]:
            if self.request.has_permission(action, item_resource):
                actions.append({
                    "name": action,
                    "title": action.capitalize(),
                    "profile": "/profiles/%s.json" % item_resource.type_info.name,
                    "href": "%s?currentAction=%s" % (
                        self.request.resource_path(item_resource), action
                    )
                })
        item["actions"] = actions
        return item

    def user_embed(self, item_id, initial_item=False):
        """
        Use request's embed method to find given item in the database.

        If the user who made the call to the API does not have
        permissions to view the item, the item will not be embedded.
        Instead, if the item is the initial ID given to the API, nothing is
        embedded; if the item is to be embedded at a subsequent depth, a
        message stating the item cannot be embedded is inserted instead.

        Additionally, if the given ID is the initial item to embed, add
        "actions" if item embedded.

        :param item_id: string uuid or @id
        :param initial_item: bool indicative of embedding initial item ID
        :return item: object to return for embedding
        """
        item = None
        given_id = item_id
        if not item_id.startswith("/"):
            item_id = "/" + item_id
        try:
            item = self.request.embed(item_id, "@@object", as_user=True)
        except HTTPForbidden:
            if not initial_item:
                item = FORBIDDEN_MSG
        except KeyError:
            self.invalid_ids.append(given_id)
        if item and initial_item:
            item = self.add_actions(item)
        return item

    def minimal_embed(self, item_id):
        """
        Embed minimal item info. Helpful for preventing recursions for
        items for which detailed info is commonly not needed.

        :param item_id: string uuid or @id
        :return item_embed: dict with item title and @id
        """
        item_object = self.user_embed(item_id)
        if item_object == FORBIDDEN_MSG:
            item_embed = item_object
        elif isinstance(item_object, dict):
            item_title = item_object.get("title", "")
            item_atid = item_object.get("@id", "")
            item_embed = {"title": item_title, "@id": item_atid}
        else:
            item_embed = item_object
        return item_embed

    @staticmethod
    def is_uuid(uuid_to_test, version=4):
        """
        Determine if given string is a valid uuid.

        :param uuid_to_test: string to check
        :param version: int for uuid version
        :return: bool if given string is valid uuid
        """
        try:
            uuid_obj = UUID(uuid_to_test, version=version)
        except ValueError:
            return False
        return str(uuid_obj) == uuid_to_test

    def embed(self, item, depth):
        """
        Embed items recursively according to input parameters. Unpack
        dictionaries and lists to find @ids, which are selectively embedded,
        typically in object view. Store new embeds in cache for look-up.

        :param item: object of interest to expand
        :param depth: int of current embed depth
        :return item: object of interest processed
        """
        while True:
            if depth == self.embed_depth:
                break
            elif isinstance(item, dict) and item:
                for key in item:
                    if key in KEYS_TO_IGNORE:
                        continue
                    item[key] = self.embed(item[key], depth)
                break
            elif isinstance(item, list) and item:
                for idx in range(len(item)):
                    item[idx] = self.embed(item[idx], depth)
                break
            elif isinstance(item, str):
                if ATID_PATTERN.match(item):
                    if depth == -1:
                        cache_item = item
                        item = self.user_embed(item, initial_item=True)
                        self.cache[cache_item] = item
                        depth += 1
                    elif self.desired_embeds:
                        if item.split("/")[1] in self.desired_embeds:
                            if item in self.cache:
                                item = self.cache[item]
                                depth += 1
                            else:
                                cache_item = item
                                item = self.user_embed(item)
                                self.cache[cache_item] = item
                                depth += 1
                        else:
                            break
                    else:
                        if item.split("/")[1] in self.ignored_embeds:
                            break
                        elif item in self.cache:
                            item = self.cache[item]
                            depth += 1
                        elif GENELIST_ATID.match(item):
                            # NOTE: Non-admins forbidden for raw view, so just skip
                            # attempt to embed gene lists for default settings.
                            break
                        elif MINIMAL_EMBED_ATID.match(item):
                            cache_item = item
                            item = self.minimal_embed(item)
                            self.cache[cache_item] = item
                            break
                        else:
                            cache_item = item
                            item = self.user_embed(item)
                            self.cache[cache_item] = item
                            depth += 1
                elif self.is_uuid(item) and depth == -1:
                    item = self.user_embed(item, initial_item=True)
                    depth += 1
                else:
                    break
            else:
                break
        return item

    def fields_to_nested_dict(self):
        """
        Convert list of requested fields into nested dictionary. Each
        nested dictionary contains keys whose values are sub-dictionaries
        to embed as well as a "fields_to_keep" key whose values are the
        terminal fields requested.

        For example, if the requested fields are:
            'variant.gene.title'
            'variant.*'
        the resulting nested dict will be:

        {
            "variant": {
                "gene": {"fields_to_keep": ["title"]},
                "fields_to_keep": ["*"]
            }
        }

        :return field_dict: nested dict of requested fields
        """
        field_dict = {}
        for field in self.requested_fields:
            field_keys = field.split(".")
            field_keys = [x for x in field_keys if x]
            field_dict = self.build_nested_dict(field_dict, field_keys)
        return field_dict

    def build_nested_dict(self, field_dict, field_keys):
        """
        Recursively builds a nested dict for each requested field by
        iterating through the keys of the requested field, adding
        the keys if not present and building a nested dict for the
        remaining keys. The terminal key of the requested field always
        corresponds to a field of an embedded item that should be included
        in the embedding.

        :param field_dict: existing dict of requested fields
        :param field_keys: list of keys of a requested field
        :return field_dict: existing dict updated with new field_keys
        """
        key = field_keys.pop(0)
        if not field_keys:
            if "fields_to_keep" in field_dict:
                field_dict["fields_to_keep"].append(key)
            else:
                field_dict["fields_to_keep"] = [key]
        else:
            if key not in field_dict:
                field_dict[key] = {}
            field_dict[key] = self.build_nested_dict(field_dict[key], field_keys)
        return field_dict

    def field_embed(self, item, field_dict, initial_item=False):
        """
        Embed items recursively according to requested fields. Follows
        keys of the nested dict of requested fields, unpacking dictionaries
        and lists and embedding @ids as required to reach the terminal
        requested fields.

        :param item: object of interest to expand
        :param field_dict: nested dict of requested fields
        :param initial_item: bool indicative of embedding initial item ID
        :return item: object of interest processed
        """
        while True:
            if isinstance(item, dict):
                if item == FORBIDDEN_MSG:
                    break
                fields_to_keep = []
                for key in field_dict:
                    if key == "fields_to_keep":
                        fields_to_keep += field_dict[key]
                        continue
                    if key not in item:
                        continue
                    fields_to_keep.append(key)
                    item[key] = self.field_embed(item[key], field_dict[key])
                if initial_item and "actions" not in fields_to_keep:
                    fields_to_keep.append("actions")
                if not initial_item and "actions" in fields_to_keep:
                    if DATABASE_ITEM_KEY in item:
                        item = self.add_actions(item)
                    else:
                        raise HTTPBadRequest(
                            "The 'actions' field was requested for a JSON object"
                            " that is not a database item."
                        )
                if "*" not in fields_to_keep:
                    culled_item = {}
                    for field in fields_to_keep:
                        try:
                            culled_item[field] = item[field]
                        except KeyError:
                            continue
                    item = culled_item
                break
            if isinstance(item, list):
                for idx in range(len(item)):
                    item[idx] = self.field_embed(item[idx], field_dict)
                break
            elif isinstance(item, str):
                if ATID_PATTERN.match(item):
                    if item in self.cache:
                        item = self.cache[item]
                    else:
                        cache_item = item
                        item = self.user_embed(item)
                        self.cache[cache_item] = item
                else:
                    break
            else:
                break
        return item


@view_config(
    route_name="embed", request_method="POST", effective_principals=Authenticated
)
@debug_log
def embed(context, request):
    """
    API to return custom-embedded view of object posted to endpoint. If no
    parameters provided, attempt to return object with embedding done
    per default parameters.

    :param context: pyramid request context
    :param request: pyramid request object
    :return results: list of dicts of custom-embedded views of items
    """
    ids = []
    ignored_embeds = []
    desired_embeds = []
    requested_fields = []
    results = []
    invalid_ids = []
    embed_depth = 4  # Arbritary standard depth to search.
    ignored(context)
    if request.GET:
        ids += request.GET.dict_of_lists().get("id", [])
        embed_depth = int(request.GET.get("depth", embed_depth))
        ignored_embeds += request.GET.dict_of_lists().get("ignored", [])
        desired_embeds += request.GET.dict_of_lists().get("desired", [])
        requested_fields += request.GET.dict_of_lists().get("field", [])
    elif request.json:
        ids += request.json.get("ids", [])
        ignored_embeds = request.json.get("ignored", [])
        desired_embeds = request.json.get("desired", [])
        embed_depth = request.json.get("depth", embed_depth)
        requested_fields = request.json.get("fields", [])
    ids = list(set(ids))
    if len(ids) > 5:
        raise HTTPBadRequest(
            "Too many items were given for embedding."
            " Please limit to less than 5 items."
        )
    if not ids:
        raise HTTPBadRequest("No item identifier was provided.")
    embed_props = {
        "ignored_embeds": ignored_embeds,
        "desired_embeds": desired_embeds,
        "embed_depth": embed_depth,
        "requested_fields": requested_fields,
    }
    for item_id in ids:
        item_embed = CustomEmbed(request, item_id, embed_props)
        results.append(item_embed.result)
        invalid_ids += item_embed.invalid_ids
    invalid_ids += [item for item in results if isinstance(item, str)]
    if invalid_ids:
        raise HTTPBadRequest(
            "The following IDs were invalid: %s" % ", ".join(invalid_ids)
        )
    return results



import os
import pytest

from dcicutils.qa_checkers import ChangeLogChecker, DebuggingArtifactChecker
from .conftest_settings import REPOSITORY_ROOT_DIR


@pytest.mark.static
def test_version_and_changelog():

    class MyAppChangeLogChecker(ChangeLogChecker):
        PYPROJECT = os.path.join(REPOSITORY_ROOT_DIR, "pyproject.toml")
        CHANGELOG = os.path.join(REPOSITORY_ROOT_DIR, "CHANGELOG.rst")

    MyAppChangeLogChecker.check_version()


@pytest.mark.static
def test_utils_debugging_artifacts_pdb():
    checker = DebuggingArtifactChecker(sources_subdir="src/encoded",
                                       skip_files="(tests/data)",
                                       filter_patterns=['pdb'])
    checker.check_for_debugging_patterns()


@pytest.mark.static
def test_utils_debugging_artifacts_print():
    checker = DebuggingArtifactChecker(sources_subdir="src/encoded",
                                       skip_files="encoded/(commands|tests)/",
                                       filter_patterns=['print'],
                                       if_used='warning')
    checker.check_for_debugging_patterns()



from collections import defaultdict
from functools import reduce
from pyramid.decorator import reify
from .interfaces import (
    CALCULATED_PROPERTIES,
    TYPES,
)
from .schema_utils import combine_schemas


def includeme(config):
    registry = config.registry
    registry[TYPES] = TypesTool(registry)


def extract_schema_links(schema):
    if not schema:
        return
    for key, prop in schema['properties'].items():
        if 'items' in prop:
            prop = prop['items']
        if 'properties' in prop:
            for path in extract_schema_links(prop):
                yield (key,) + path
        elif 'linkTo' in prop:
            yield (key,)


class AbstractTypeInfo(object):
    """
    Contains meta information per item type held in a AbstractCollection.
    Does not actually use the item resource itself. Has properties for finding
    subtypes and child types of the given item, as well as a schema that is
    formed by merging the subschemas.
    """
    factory = None

    def __init__(self, registry, name):
        """
        Args:
            registry (Registry): the Pyramid registry
            name (str): item name, e.g. "MyItem"
        """
        self.types = registry[TYPES]
        self.name = name

    @reify
    def subtypes(self):
        return [
            ti.name for ti in self.types.by_item_type.values()
            if self.name in ([ti.name] + ti.base_types)
        ]

    @reify
    def child_types(self):
        """
        Need to handle non-abstract item types and abstract types separately,
        due to the pre-existing setup of base_types
        """
        child_types = [
            ti.name for ti in self.types.by_item_type.values()
            if self.name == ti.base_types[0]
        ]
        # child abstract types include the parent name, but not the type itself
        child_types.extend([
            ti.name for ti in self.types.by_abstract_type.values()
            if (self.name != ti.name and self.name in ti.base_types)
        ])
        return child_types

    @reify
    def schema(self):
        """
        Schema resulting from merging the subschemas. NOT equivalent to the
        schema defined for the item type itself (use the TypeInfo contained in
        registry[TYPES].by_abstract_type[item_type] for that)
        """
        subschemas = (self.types[name].schema for name in self.subtypes)
        return reduce(combine_schemas, subschemas)


class TypeInfo(AbstractTypeInfo):
    """
    Extends AbstractTypeInfo for meta information per item type that is held
    in a Collection. Has properties to reference the schema and various
    related attributes, as well as connections to properties on the item
    itself (through `factory`)
    """
    def __init__(self, registry, item_type, factory, abstract=False):
        """
        Args:
            registry (Registry): the Pyramid registry
            item_type (str): item type for the item, e.g. "my_item"
            factory (Item): actual resource for the item, e.g. snovault.resources.Item
            abstract (bool): used to keep track if this is used for an abstract item
        """
        super(TypeInfo, self).__init__(registry, factory.__name__)
        self.registry = registry
        self.item_type = item_type
        self.factory = factory
        self.base_types = factory.base_types
        self.aggregated_items = factory.aggregated_items
        self.embedded_list = factory.embedded_list
        self.default_diff = factory.default_diff
        self.is_abstract = abstract

    @reify
    def calculated_properties(self):
        return self.registry[CALCULATED_PROPERTIES]

    @reify
    def schema_version(self):
        try:
            return self.factory.schema['properties']['schema_version']['default']
        except (KeyError, TypeError):
            return None

    @reify
    def schema_links(self):
        return sorted('.'.join(path) for path in extract_schema_links(self.factory.schema))

    @reify
    def schema_keys(self):
        if not self.factory.schema:
            return ()
        keys = defaultdict(list)
        for key, prop in self.factory.schema['properties'].items():
            uniqueKey = prop.get('items', prop).get('uniqueKey')
            if uniqueKey is True:
                uniqueKey = '%s:%s' % (self.factory.item_type, key)
            if uniqueKey is not None:
                keys[uniqueKey].append(key)
        return keys

    @reify
    def merged_back_rev(self):
        merged = {}
        types = [self.name] + self.base_types
        for name in reversed(types):
            back_rev = self.types.type_back_rev.get(name, ())
            merged.update(back_rev)
        return merged

    @reify
    def schema(self):
        props = self.calculated_properties.props_for(self.factory)
        schema = self.factory.schema or {'type': 'object', 'properties': {}}
        schema = schema.copy()
        schema['properties'] = schema['properties'].copy()
        for name, prop in props.items():
            if prop.schema is not None:
                schema['properties'][name] = prop.schema
        return schema


class TypesTool(object):
    """
    Helper class used to register and store TypeInfo/AbstractTypeInfo classes
    corresponding to different item types.
    Whether an item type is abstract or not depends on the collection decorator
    used for the function; see snovault.config.py for more information.
    Below will refer to registry[TYPES] as "types"...

    Includes:
    - TypeInfo objects (for Collections) that are registered in types.all by
      item type (e.g. my_item), item name (e.g. MyItem), and item class; also
      types.by_item_type using item_type. Uses `register` method.
    - AbstractTypeInfo objects (for AbstractCollections) that are registed in
      types.all by item name and item class. Additionally, register the
      TypeInfo for the item in types.by_abstract_type using item_type.
      Uses the `register_abstract` method.
    """
    def __init__(self, registry):
        self.registry = registry
        self.by_item_type = {}
        self.by_abstract_type = {}
        self.type_back_rev = {}
        self.all = {}

    def register(self, factory):
        name = factory.__name__
        item_type = factory.item_type or name
        ti = TypeInfo(self.registry, item_type, factory)
        self.all[ti.item_type] = self.by_item_type[ti.item_type] = ti
        self.all[ti.name] = ti
        self.all[ti.factory] = ti
        # for base in ti.base_types:
        #     self.register_abstract(base)

        # Calculate the reverse rev map
        for prop_name, spec in factory.rev.items():
            rev_type_name, rel = spec
            back = self.type_back_rev.setdefault(rev_type_name, {}).setdefault(rel, set())
            back.add((ti.name, prop_name))

        return ti

    def register_abstract(self, factory):
        # create the TypeInfo and register in self.by_abstract_type
        # `item_type` is likely not set on the Items of abstract collections
        name = factory.__name__
        item_type = factory.item_type or name
        ti = TypeInfo(self.registry, item_type, factory, abstract=True)
        self.by_abstract_type[ti.item_type] = ti

        # now create the AbstractTypeInfo, which is also registered
        abstract_ti = AbstractTypeInfo(self.registry, name)
        self.all[factory] = self.all[name] = abstract_ti
        return abstract_ti

    def __contains__(self, name):
        return name in self.all

    def __getitem__(self, name):
        return self.all[name]



import json

from dcicutils.misc_utils import environ_bool, PRINT
from . import COLLECTIONS
from pyramid.security import Authenticated


DEBUG_PERMISSIONS = environ_bool("DEBUG_PERMISSIONS", default=False)


def groupfinder(login, request):
    if '.' not in login:
        if DEBUG_PERMISSIONS:
            PRINT("groupfinder sees no '.' in %s, returning None" % login)
        return None
    namespace, localname = login.split('.', 1)
    user = None

    collections = request.registry[COLLECTIONS]

    """ At least part of this stanza seems mainly for testing purposes
        should the testing bits be refactored elsewhere???
        20-09-08 changed permission model requires import of Authenticated
        is that kosher
    """
    # TODO (C4-332): Consolidate permissions all in one perms.py file once this all stabilizes.
    if namespace == 'remoteuser':

        # These names are used in testing or special service situations to force the permissions result
        # to known values without any need to go through lookup of any particular user and process
        # their groups or project_roles.

        synthetic_result = None

        if localname in ['EMBED', 'INDEXER']:
            synthetic_result = []
        elif localname in ['TEST', 'IMPORT', 'UPGRADE', 'INGESTION']:
            synthetic_result = ['group.admin']
        elif localname in ['TEST_SUBMITTER']:
            synthetic_result = ['group.submitter']
        elif localname in ['TEST_AUTHENTICATED']:
            synthetic_result = [Authenticated]

        if synthetic_result is not None:
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "returning synthetic result:", synthetic_result)
            return synthetic_result

        # Note that the above 'if' has no final 'else', and the remainder of cases,
        # having the form remoteuser.<username>, are processed in the next 'if' below.

    if namespace in ('mailto', 'remoteuser', 'auth0'):
        users = collections.by_item_type['user']
        try:
            user = users[localname]
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "found user", localname)
        except KeyError:
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "failed to find user", localname)
            return None

    elif namespace == 'accesskey':

        access_keys = collections.by_item_type['access_key']
        try:
            access_key = access_keys[localname]
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "found access key", localname)
        except KeyError:
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "failed to find access key", localname)
            return None

        access_key_status = access_key.properties.get('status')
        if access_key_status in ('deleted', 'revoked'):
            if DEBUG_PERMISSIONS:
                PRINT("groupfinder for", login, "found", access_key_status, "access key", localname)
            return None

        userid = access_key.properties['user']
        user = collections.by_item_type['user'][userid]

        if DEBUG_PERMISSIONS:
            PRINT("groupfinder for", login, "decoded access key", localname, "as user", user)

    if user is None:
        PRINT("groupfinder for", login, "returning None because user is None")
        return None

    user_properties = user.properties

    if user_properties.get('status') in ('deleted'):
        if DEBUG_PERMISSIONS:
            PRINT("groupfinder for %s found user %s, but that user has status deleted." % (login, user))
        return None

    principals = ['userid.%s' % user.uuid]
    if DEBUG_PERMISSIONS:
        PRINT("groupfinder starting with principals", principals)

    def add_principal(principal):
        if DEBUG_PERMISSIONS:
            PRINT("groupfinder for", login, "adding", principal , "to principals.")
        principals.append(principal)

    for group in user_properties.get('groups', []):
        add_principal('group.%s' % group)

    if DEBUG_PERMISSIONS:
        PRINT("groupfinder for", login, "returning principals", json.dumps(principals, indent=2))

    return principals


def is_admin_request(request):
    """ Checks for 'group.admin' in effective_principals on request - if present we know this
        request was submitted by an admin
    """
    return 'group.admin' in request.effective_principals



import pytest
from .test_attachment import testing_download  # noQA fixture import
from ..drs import REQUIRED_FIELDS


class TestDRSAPI:
    """ Class for testing the DRS implementation - uses TestingDownload as it implements
        the @@download scheme
    """

    def test_drs_get_object(self, testapp, testing_download):  # noQA fixture
        """ Tests basic structure about a drs object """
        res = testapp.get(testing_download)
        drs_object_uri = res.json['uuid']
        drs_object_1 = testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}').json
        for key in REQUIRED_FIELDS:
            assert key in drs_object_1
        assert drs_object_1['self_uri'] == f'drs://localhost:80/ga4gh/drs/v1/objects/{drs_object_uri}'
        assert (drs_object_1['access_methods'][0]['access_url']['url']
                == f'http://localhost:80/{drs_object_uri}/@@download')

        # failure cases
        testapp.get(f'/ga4gh/drs/v1/objects/not_a_uri', status=404)

        # @@drs case
        drs_object_2 = testapp.get(f'/{drs_object_uri}/@@drs')
        for key in REQUIRED_FIELDS:
            assert key in drs_object_2

    def test_drs_get_object_url(self, testapp, testing_download):  # noQA fixture
        """ Tests extracting URL through ga4gh pathway """
        res = testapp.get(testing_download)
        drs_object_uri = res.json['uuid']

        # standard URI with meaningful access_id, discarded
        drs_object_download = testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}/access/https').json
        assert drs_object_download == {
            'url': f'http://localhost:80/{drs_object_uri}/@@download'
        }

        # /access/ method
        drs_object_download = testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}/access/').json
        assert drs_object_download == {
            'url': f'http://localhost:80/{drs_object_uri}/@@download'
        }

        # standard URI with nonsense access id, still discarded
        drs_object_download = testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}/access/blah').json
        assert drs_object_download == {
            'url': f'http://localhost:80/{drs_object_uri}/@@download'
        }

        # /access method
        drs_object_download = testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}/access').json
        assert drs_object_download == {
            'url': f'http://localhost:80/{drs_object_uri}/@@download'
        }

    def test_drs_get_object_failure(self, testapp, testing_download):  # noQA fixture
        """ Tests a bunch of bunk URLs """
        res = testapp.get(testing_download)
        drs_object_uri = res.json['uuid']

        with pytest.raises(Exception):
            testapp.get(f'/ga4gh/drs/v1/objects/not_a_uri/access/https')
        with pytest.raises(Exception):
            testapp.get(f'/ga4gh/drs/v1/objects/access/https')
        with pytest.raises(Exception):
            testapp.get(f'/ga4gh/drs/v1/objects/access/')
        with pytest.raises(Exception):
            testapp.get(f'/ga4gh/drs/v1/objects/access')
        with pytest.raises(Exception):
            testapp.get(f'/ga4gh/drs/v1/objects/{drs_object_uri}/accesss/https')



import os
import toml

from pkg_resources import resource_filename


# This isn't the home of snovault, but the home of the snovault-based application.
# So in CGAP, for example, this would want to be the home of the CGAP application.
# If not set, it will be assumed that the current working directory is that.
APPLICATION_PROJECT_HOME = os.environ.get("APPLICATION_PROJECT_HOME", os.path.abspath(os.curdir))
PYPROJECT_TOML_FILE = os.path.join(APPLICATION_PROJECT_HOME, "pyproject.toml")
PYPROJECT_TOML = toml.load(PYPROJECT_TOML_FILE)
POETRY_DATA = PYPROJECT_TOML['tool']['poetry']

PROJECT_NAME = POETRY_DATA['name'].replace('dcic', '')

print("=" * 80)
print(f"APPLICATION_PROJECT_HOME={APPLICATION_PROJECT_HOME}")
print(f"PYPROJECT_TOML_FILE={PYPROJECT_TOML_FILE}")
print(f"PROJECT_NAME={PROJECT_NAME}")
print("=" * 80)


def project_filename(filename):
    # TODO: In fact we should do this based on the working dir so that when this is imported to another repo,
    #       it gets the inserts out of that repo's tests, not our own.
    return resource_filename(PROJECT_NAME, filename)





import json
import logging
import os

import psutil
import time

from dcicutils.misc_utils import environ_bool, PRINT, ignored
from functools import lru_cache
from pkg_resources import resource_filename
from pyramid.events import BeforeRender, subscriber
from pyramid.httpexceptions import (
    HTTPMovedPermanently,
    HTTPPreconditionFailed,
    HTTPUnauthorized,
    HTTPUnsupportedMediaType,
    HTTPNotAcceptable,
    HTTPServerError
)
from pyramid.response import Response
from pyramid.settings import asbool
from pyramid.threadlocal import manager
from pyramid.traversal import split_path_info, _join_path_tuple
from subprocess_middleware.worker import TransformWorker
from urllib.parse import urlencode
from webob.cookies import Cookie
from .util import content_type_allowed


log = logging.getLogger(__name__)


def includeme(config):
    """
    Can get tween ordering by executing the following on command-line from root dir:
        `bin/ptween development.ini`

    We could alternatively put these in the base.ini file explicitly.

    See: https://docs.pylonsproject.org/projects/pyramid/en/latest/narr/hooks.html#registering-tweens

    --- IMPORTANT ---
    The `handler` arg of 'tween factory' functions refers to the subsequent tween to be called.
    This means that if handler(request) is called, then the downstream tweens are acted upon it,
    until response is returned. It's an ONION depending on where handler(request) called within a tween
    and NOT necessarily an explicit ordering (unless `return handler(request)` is last line of each tween).

    A request goes down the tween chain from INGRESS to MAIN and then back up to INGRESS.
    `handler(request)` calls the subsequent tween and returns complete tweened-from-there response.

    Tween Chain as of 05/23/2019:

        Position    Name
        --------    ----
        -           INGRESS
        0           snovault.stats.stats_tween_factory
        1           .renderers.validate_request_tween_factory
        2           .renderers.render_page_html_tween_factory
        3           .renderers.set_response_headers_tween_factory
        4           pyramid_tm.tm_tween_factory
        5           .renderers.security_tween_factory
        6           pyramid.tweens.excview_tween_factory
        -           MAIN

    The `handler` kwarg of tween factories refers to the subsequent tween to be called.
    This means that if handler(request) is called, then the downstream tweens are acted upon it,
    until response is returned. It's an ONION!

    """

    config.add_tween('.renderers.validate_request_tween_factory', under='snovault.stats.stats_tween_factory')
    # DISABLED - .add_tween('.renderers.remove_expired_session_cookies_tween_factory',
    #                       under='.renderers.validate_request_tween_factory')
    config.add_tween('.renderers.render_page_html_tween_factory', under='.renderers.validate_request_tween_factory')

    # The above tweens, when using response (= `handler(request)`) act on the _transformed_ response
    # (containing HTML body).
    # The below tweens run _before_ the JS rendering. Responses in these tweens have not been transformed to HTML yet.
    config.add_tween('.renderers.set_response_headers_tween_factory', under='.renderers.render_page_html_tween_factory')

    # If this isn't placed under 'pyramid_tm.tm_tween_factory' (e.g. under resp headers or something)
    # then the headers aren't preserved or available in server-side render or response.
    config.add_tween('.renderers.security_tween_factory', under='pyramid_tm.tm_tween_factory')

    config.scan(__name__)


def validate_request_tween_factory(handler, registry):
    """
    Updates request.environ's REQUEST_METHOD to be X_REQUEST_METHOD if present.
    Asserts that if a POST (or similar) request is in application/json format,
    with exception for /metadata/* endpoints.

    Apache config:
        SetEnvIf Request_Method HEAD X_REQUEST_METHOD=HEAD
    """
    ignored(registry)

    def validate_request_tween(request):

        # Fix Request method changed by mod_wsgi.
        # See: https://github.com/GrahamDumpleton/mod_wsgi/issues/2
        environ = request.environ
        if 'X_REQUEST_METHOD' in environ:
            environ['REQUEST_METHOD'] = environ['X_REQUEST_METHOD']

        if request.method in ('GET', 'HEAD'):
            # If GET request, don't need to check `request.content_type`
            # Includes page text/html requests.
            return handler(request)

        elif content_type_allowed(request):
            return handler(request)

        else:
            detail = "Request content type %s is not 'application/json'" % request.content_type
            raise HTTPUnsupportedMediaType(detail)

    return validate_request_tween


def security_tween_factory(handler, registry):
    ignored(registry)

    def security_tween(request):
        """
        Executed inside/prior-to any page transforms and inside/prior-to
        `pyramid_tm.tm_tween_factory` (transaction management tween).
        This is because request.authenticated_userid as well as `request.user_info`
        property getters _may_ access Postgres DB to get user properties (if not yet
        indexed in ES) and all DB transactions must complete before transaction
        management tween is completed.
        """

        expected_user = request.headers.get('X-If-Match-User')
        if expected_user is not None:  # Not sure when this is the case
            if request.authenticated_userid != 'mailto.' + expected_user:
                detail = 'X-If-Match-User does not match'
                raise HTTPPreconditionFailed(detail)
        if request.authorization is not None or asbool(request.headers.get('X-Auth-Challenge', False)):
            # wget may only send credentials following a challenge response.
            if not request.authenticated_userid:
                if not hasattr(request, 'auth0_expired'):
                    # Not a "Bearer" JWT token in Auth header. Or other error.
                    # We send a 401 "Unauthorized" exception if authentication issue or expiration.
                    # We send a 403 "Forbidden" (TODO: assert) if authorized correctly but no view permission
                    raise HTTPUnauthorized(
                        title="No Access",
                        comment="Invalid Authorization header or Auth Challenge response.",
                        headers={
                            'WWW-Authenticate': (
                                f'Bearer realm="{request.domain}";'
                                f' Basic realm="{request.domain}"'
                            )
                        }
                    )

        if hasattr(request, 'auth0_expired'):
            # Add some security-related headers on the up-swing
            response = handler(request)
            if request.auth0_expired:
                # return response
                #
                # If have the attribute and it is true, then our session has expired.
                # This is true for both AJAX requests (which have request.authorization) & browser page
                # requests (which have cookie); both cases handled in authentication.py
                # Informs client or libs/react-middleware.js serverside render of expired token
                # to set logged-out state in front-end in either doc request or xhr request & set appropriate alerts
                response.headers['X-Request-JWT'] = "expired"

                # Especially for initial document requests by browser, but also desired for AJAX and other requests,
                # unset jwtToken cookie so initial client-side React render has App(instance).state.session = false
                # to be synced w/ server-side
                response.set_cookie(
                    name='jwtToken',
                    value=None,  # = i.e., same as response.delete_cookie(..)
                    domain=request.domain,
                    max_age=0,
                    path='/',
                    overwrite=True
                )
                response.status_code = 401
                response.headers['WWW-Authenticate'] = (
                    f'Bearer realm="{request.domain}", title="Session Expired";'
                    f' Basic realm="{request.domain}"'
                )
            else:
                # We have JWT and it's not expired. Add 'X-Request-JWT' & 'X-User-Info' header.
                # For performance, only do it if should transform to HTML as is not needed on every request.
                if should_transform(request, response):
                    login = request.authenticated_userid
                    if login:
                        authtype, email = login.split('.', 1)
                        if authtype == 'auth0':
                            # This header is parsed in renderer.js, or, more accurately,
                            # by libs/react-middleware.js which is imported by server.js and compiled into
                            # renderer.js. Is used to get access to User Info on initial web page render.
                            response.headers['X-Request-JWT'] = request.cookies.get('jwtToken', '')
                            user_info = request.user_info.copy()  # Re-ified property set in authentication.py
                            response.headers['X-User-Info'] = json.dumps(user_info)
                        else:
                            response.headers['X-Request-JWT'] = "null"
            return response

        return handler(request)

    return security_tween


def remove_expired_session_cookies_tween_factory(handler, registry):
    """
    CURRENTLY DISABLED
    Original purpose of this was to remove expired (session?) cookies.
    See: https://github.com/ENCODE-DCC/encoded/commit/75854803c99e5044a6a33aedb3a79d750481b6cd#diff-bc19a9793a1b3b4870cff50e7c7c9bd1R135

    We disable it for now via removing from tween chain as are using JWT tokens and handling
    their removal in security_tween_factory & authentication.py as well as client-side
    (upon "Logout" action). If needed for some reason, can re-enable.
    """  # noQA - not going to break the long URL line above
    ignored(registry)

    ignore = {
        '/favicon.ico',
    }

    def remove_expired_session_cookies_tween(request):
        if request.path in ignore or request.path.startswith('/static/'):
            return handler(request)

        session = request.session
        # if session or session._cookie_name not in request.cookies:
        #     return handler(request)

        response = handler(request)
        # Below seems to be empty always; though we do have some in request.cookies
        existing = response.headers.getall('Set-Cookie')
        if existing:
            cookies = Cookie()
            for header in existing:
                cookies.load(header)
            if session._cookie_name in cookies:
                return response

        response.delete_cookie(
            session._cookie_name,
            path=session._cookie_path,
            domain=session._cookie_domain,
        )

        return response

    return remove_expired_session_cookies_tween


def set_response_headers_tween_factory(handler, registry):
    """Add additional response headers here"""
    ignored(registry)

    def set_response_headers_tween(request):
        response = handler(request)
        response.headers['X-Request-URL'] = request.url
        # Setter automatically converts set back to tuple.
        # See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Vary
        response.vary = set((response.vary or ()) + ('Accept', 'Authorization'))
        return response

    return set_response_headers_tween


@subscriber(BeforeRender)
def canonical_redirect(event):
    request = event['request']

    # Ignore subrequests
    if len(manager.stack) > 1:
        return

    if request.method not in ('GET', 'HEAD'):
        return
    if request.response.status_int != 200:
        return
    if not request.environ.get('encoded.canonical_redirect', True):
        return
    if request.path_info == '/':
        return

    if not isinstance(event.rendering_val, dict):
        return

    canonical = event.rendering_val.get('@id', None)
    if canonical is None:
        return
    canonical_path, _, canonical_qs = canonical.partition('?')

    request_path = _join_path_tuple(('',) + split_path_info(request.path_info))
    if (request_path == canonical_path.rstrip('/') and
            request.path_info.endswith('/') == canonical_path.endswith('/') and
            (canonical_qs in ('', request.query_string))):
        return

    if '/@@' in request.path_info:
        return

    qs = canonical_qs or request.query_string
    # add redirect information to the query string, but not for the routes specified below
    if not any(route in canonical_path for route in ['/search/', '/browse/', '/metadata/']):
        redir_qs = (qs + '&' if qs else '') + urlencode([('redirected_from', request.path_info)])
    else:
        redir_qs = qs
    location = canonical_path + ('?' if redir_qs else '') + redir_qs
    raise HTTPMovedPermanently(location=location, detail="Redirected from " + str(request.path_info))


# Web browsers send an Accept request header for initial (e.g. non-AJAX) page requests
# which should contain 'text/html'
MIME_TYPE_HTML = 'text/html'
MIME_TYPE_JSON = 'application/json'
MIME_TYPE_LD_JSON = 'application/ld+json'

# Note: In cgap-portal, MIME_TYPE_JSON is at the head of this list. In fourfront, MIME_TYPE_HTML is.
# The cgap-portal behavior might be a bug we should look at bringing into alignment. -kmp 29-Jan-2022
MIME_TYPES_SUPPORTED = [MIME_TYPE_JSON, MIME_TYPE_HTML, MIME_TYPE_LD_JSON]
MIME_TYPE_DEFAULT = MIME_TYPES_SUPPORTED[0]
MIME_TYPE_TRIAGE_MODE = 'modern'  # if this doesn't work, fall back to 'legacy'

DEBUG_MIME_TYPES = environ_bool("DEBUG_MIME_TYPES", default=False)


def best_mime_type(request, mode=MIME_TYPE_TRIAGE_MODE):
    # TODO: I think this function does nothing but return MIME_TYPES_SUPPORTED[0] -kmp 3-Feb-2021
    """
    Given a request, tries to figure out the best kind of MIME type to use in response
    based on what kinds of responses we support and what was requested.

    In the case we can't comply, we just use application/json whether or not that's what was asked for.
    """
    if mode == 'legacy':
        # See:
        # https://tedboy.github.io/flask/generated/generated/werkzeug.Accept.best_match.html#werkzeug-accept-best-match
        # Note that this is now deprecated, or will be. The message is oddly worded ("will be deprecated")
        # that presumably means "will be removed". Deprecation IS the warning of actual action, not the action itself.
        # "This is currently maintained for backward compatibility, and will be deprecated in the future.
        #  AcceptValidHeader.best_match() uses its own algorithm (one not specified in RFC 7231) to determine
        #  what is a best match. The algorithm has many issues, and does not conform to RFC 7231."
        # Anyway, we were getting this warning during testing:
        #   DeprecationWarning: The behavior of AcceptValidHeader.best_match is currently
        #      being maintained for backward compatibility, but it will be deprecated in the future,
        #      as it does not conform to the RFC.
        # TODO: Once the modern replacement is shown to work, we should remove this conditional branch.
        result = request.accept.best_match(MIME_TYPES_SUPPORTED, MIME_TYPE_DEFAULT)
    else:
        options = request.accept.acceptable_offers(MIME_TYPES_SUPPORTED)
        if not options:
            # TODO: Probably we should return a 406 response by raising HTTPNotAcceptable if
            #       no acceptable types are available. (Certainly returning JSON in this case is
            #       not some kind of friendly help toa naive user with an old browser.)
            #       Ref: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
            result = MIME_TYPE_DEFAULT
        else:
            mime_type, score = options[0]
            result = mime_type
    if DEBUG_MIME_TYPES:
        PRINT("Using mime type", result, "for", request.method, request.url)
        for k, v in request.headers.items():
            PRINT("%s: %s" % (k, v))
        PRINT("----------")
    return result


@lru_cache(maxsize=16)
def should_transform(request, response):
    """
    Determines whether to transform the response from JSON->HTML/JS depending on type of response
    and what the request is looking for to be returned via these criteria, which are tried in order
    until one succeeds:

    * If the request method is other than GET or HEAD, returns False.
    * If the response.content_type is other than 'application/json', returns False.
    * If a 'frame=' query param is given and not 'page' (the default), returns False.
    * If a 'format=json' query param is given explicitly,
        * For 'format=html', returns True.
        * For 'format=json', returns False.
      This rule does not match if 'format=' is not given explicitly.
      If 'format=' is given an explicit value of ther than 'html' or 'json', an HTTPNotAcceptable error will be raised.
    * If the first element of MIME_TYPES_SUPPORTED[0] is 'text/html', returns True.
    * Otherwise, in all remaining cases, returns False.

    NOTE: Memoized via `lru_cache`. Cache size is set to be 16 (> 1) in case sub-requests fired off during handling.
    """
    # We always return JSON in response to POST, PATCH, etc.
    if request.method not in ('GET', 'HEAD'):
        return False

    # Only JSON response/content can be plugged into HTML/JS template responses.
    if response.content_type != 'application/json':
        return False

    # If we have a 'frame' that is not None or page, force JSON, since our UI doesn't handle all various
    # forms of the data, just embedded/page.
    request_frame = request.params.get("frame", "page")
    if request_frame != "page":
        return False

    # The `format` URI param allows us to override request's 'Accept' header.
    format_param = request.params.get('format')
    if format_param is not None:
        format_param = format_param.lower()
        if format_param == 'json':
            return False
        if format_param == 'html':
            return True
        else:
            raise HTTPNotAcceptable("Improper format URI parameter",
                                    comment="The format URI parameter should be set to either html or json.")

    # Web browsers send an Accept request header for initial (e.g. non-AJAX) page requests
    # which should contain 'text/html'
    # See: https://tedboy.github.io/flask/generated/generated/werkzeug.Accept.best_match.html#werkzeug-accept-best-match
    mime_type = best_mime_type(request)  # Result will be one of MIME_TYPES_SUPPORTED

    # N.B. ld+json (JSON-LD) is likely more unique case and might be sent by search engines (?)
    # which can parse JSON-LDs. At some point we could maybe have it to be same as
    # making an `@@object` or `?frame=object` request (?) esp if fill
    # out @context response w/ schema(s) (or link to schema)

    return mime_type == MIME_TYPE_HTML


def render_page_html_tween_factory(handler, registry):

    class TransformErrorResponse(HTTPServerError):
        """Extends 500 server error"""
        explanation = 'Transformation of JSON response to HTML webpage failed.'

    node_env = os.environ.copy()
    node_env['NODE_PATH'] = ''

    # Rendering huge pages can make the node process memory usage explode.
    # Ideally we would let the OS handle this with `ulimit` or by calling
    # `resource.setrlimit()` from  a `subprocess.Popen(preexec_fn=...)`.
    # Unfortunately Linux does not enforce RLIMIT_RSS.
    # An alternative would be to use cgroups, but that makes per-process limits
    # tricky to enforce (we would need to create one cgroup per process.)
    # So we just manually check the resource usage after each transform.

    rss_limit = 512 * (1024 ** 2)  # MB

    reload_process = (True
                      if registry.settings.get('reload_templates', False)
                      else lambda proc: psutil.Process(proc.pid).memory_info().rss > rss_limit)

    # TransformWorker inits and manages a subprocess
    # it re-uses the subprocess so interestingly data in JS global variables
    # might persist in between renders (from different users, even).
    transform = TransformWorker(
        Response=Response,
        reload_process=reload_process,
        # Other kwargs, including env below, get passed down to subprocess.Popen
        args=['node', resource_filename(__name__, 'static/build/renderer.js')],
        env=node_env
    )

    def render_page_html_tween(request):
        # Result of downstream tweens. Body not yet transformed into HTML.
        response = handler(request)

        if not should_transform(request, response):
            # Continue back up the tween chain with JSON response body.
            return response

        # The stats below are converted into "X-Stats" header in snovault.
        # Maybe we could conditionally disable this at some point in .ini config
        # for minute performance enhancement(s).

        transform_start_time = time.time()

        try:
            response = transform(response)
        except ValueError as e:
            response = TransformErrorResponse(e.args[0])
        else:
            transform_end_time = time.time()
            transform_duration = int((transform_end_time - transform_start_time) * 1e6)
            stats = request._stats
            stats['render_count'] = stats.get('render_count', 0) + 1
            stats['render_time'] = stats.get('render_time', 0) + transform_duration
            # We don't care about getting this back in form of a cookie. Will be available
            # as header.
            request._add_stats_cookie = False

        return response

    return render_page_html_tween



import contextlib
import transaction

from snovault import DBSESSION
from snovault.storage import Base
from sqlalchemy import MetaData
from zope.sqlalchemy import mark_changed


# Once debugged, this support probably wants to move to snovault.

class PyramidAppManager:

    def __init__(self, app):
        self.session = app.registry[DBSESSION]
        self._meta = None
        self._ordered_table_names = None

    def _reflect(self):
        if self._meta is None:
            meta = MetaData(bind=self.session.connection())
            meta.reflect()
            self._meta = meta

    @property
    def meta(self):
        self._reflect()
        return self._meta

    @property
    def ordered_table_names(self):
        ordered_names = self._ordered_table_names
        if ordered_names is None:
            self._reflect()
            self._ordered_table_names = ordered_names = reversed(Base.metadata.sorted_tables)
        return ordered_names

    @contextlib.contextmanager
    def connection(self, as_transaction=False):
        """
        Context manager executes a body of code with a connection object to the database.

        :param as_transaction: If the action is expected to be read-only, this can be false.
            If there will be modifications that need to be committed, specify as_transaction=True.

        """
        connection = self.session.connection().connect()
        if not transaction:
            yield connection
        else:
            try:
                yield connection
            except Exception:
                transaction.abort()
            else:
                # commit all changes to DB
                self.session.flush()
                mark_changed(self.session())
                transaction.commit()



import json

from dcicutils.misc_utils import exported
from dcicutils.es_utils import create_es_client
from elasticsearch.connection import RequestsHttpConnection
from elasticsearch.serializer import SerializationError
from pyramid.settings import asbool
from ..json_renderer import json_renderer
from ..util import get_root_request
from .interfaces import APP_FACTORY, ELASTIC_SEARCH


exported(
    APP_FACTORY,  # e.g., re-exported by snovault/__init__.py
    ELASTIC_SEARCH,  # e.g., imported by snovault/tests/search.py
)


def includeme(config):
    """ Creates the es_client for use by the application
        Important options from settings:
            * elasticsearch.server - URL of the server
            * elasticsearch.aws_auth - whether or not to use local AWS creds
            * elasticsearch.request_timeout - ES request timeout, defaults to 10 seconds but is
              upped to 20 in our settings if not otherwise specified.
            * elasticsearch.request_auto_retry - allows us to disable request auto-retry
    """
    settings = config.registry.settings
    address = settings['elasticsearch.server']
    use_aws_auth = settings.get('elasticsearch.aws_auth')
    es_request_timeout = settings.get('elasticsearch.request_timeout', 20)  # only change is here
    es_request_auto_retry = settings.get('elasticsearch.request_auto_retry', True)
    # make sure use_aws_auth is bool
    if not isinstance(use_aws_auth, bool):
        use_aws_auth = True if use_aws_auth == 'true' else False
    # snovault specific ES options
    # this previously-used option was causing problems (?)
    # 'connection_class': TimedUrllib3HttpConnection
    es_options = {'serializer': PyramidJSONSerializer(json_renderer),
                  'connection_class': TimedRequestsHttpConnection,
                  'timeout': es_request_timeout,
                  'retry_on_timeout': es_request_auto_retry}

    config.registry[ELASTIC_SEARCH] = create_es_client(address,
                                                       use_aws_auth=use_aws_auth,
                                                       **es_options)

    config.include('.cached_views')
    config.include('.esstorage')
    config.include('.indexer_queue')
    config.include('.indexer_utils')  # has invalidation_scope route
    config.include('.indexer')
    if asbool(settings.get('mpindexer')):
        config.include('.mpindexer')


class PyramidJSONSerializer(object):
    mimetype = 'application/json'

    def __init__(self, renderer):
        self.renderer = renderer

    @staticmethod
    def loads(s):
        try:
            return json.loads(s)
        except (ValueError, TypeError) as e:
            raise SerializationError(s, e)

    def dumps(self, data):
        # don't serialize strings
        if isinstance(data, (type(''), type(u''))):
            return data

        try:
            return self.renderer.dumps(data)
        except (ValueError, TypeError) as e:
            raise SerializationError(data, e)


# changed to work with Urllib3HttpConnection (from ES) to RequestsHttpConnection
class TimedRequestsHttpConnection(RequestsHttpConnection):
    stats_count_key = 'es_count'
    stats_time_key = 'es_time'

    def stats_record(self, duration):
        request = get_root_request()
        if request is None:
            return

        duration = int(duration * 1e6)
        if not hasattr(request, "_stats"):
            request._stats = {}
        stats = request._stats
        stats[self.stats_count_key] = stats.get(self.stats_count_key, 0) + 1
        stats[self.stats_time_key] = stats.get(self.stats_time_key, 0) + duration

    def log_request_success(self, method, full_url, path, body, status_code, response, duration):
        self.stats_record(duration)
        return super(TimedRequestsHttpConnection, self).log_request_success(
            method, full_url, path, body, status_code, response, duration)

    def log_request_fail(self, method, full_url, path, body, duration, status_code=None, response=None, exception=None):
        self.stats_record(duration)
        return super(TimedRequestsHttpConnection, self).log_request_fail(
            method, full_url, path, body, duration, status_code=status_code, response=response, exception=exception)



import structlog
from elasticsearch.helpers import scan
from pyramid.view import view_config
from pyramid.exceptions import HTTPBadRequest
from ..interfaces import COLLECTIONS, TYPES
from .interfaces import ELASTIC_SEARCH
from ..util import DEFAULT_EMBEDS, crawl_schema, debug_log
from ..typeinfo import AbstractTypeInfo


log = structlog.getLogger(__name__)
SCAN_PAGE_SIZE = 5000


def includeme(config):
    config.add_route('compute_invalidation_scope', '/compute_invalidation_scope')
    config.scan(__name__)


def get_namespaced_index(config, index):
    """ Grabs indexer.namespace from settings and namespace the given index """
    try:
        settings = config.registry.settings
    except AttributeError:  # accept either config or registry as first arg
        settings = config.settings
    namespace = settings.get('indexer.namespace') or ''
    return namespace + index


def namespace_index_from_health(health, index):
    """ Namespaces the given index based on health page data """
    if 'error' in health:
        raise RuntimeError('Mirror health unresolved: %s' % health)
    return health.get('namespace', '') + index


def to_camel_case(snake_string):
    return snake_string.title().replace("_", "")


def find_uuids_for_indexing(registry, updated, find_index=None):
    """
    Run a search to find uuids of objects with that contain the given set of
    updated uuids in their linked_uuids.
    Uses elasticsearch.helpers.scan to iterate through ES results.
    Returns a set containing original uuids and the found uuids (INCLUDING
    uuids that were passed into this function)

    Args:
        registry: the current Registry
        updated (set): uuids to use as basis for finding associated items
        find_index (str): index to search in. Default to '_all' (all indices)

    Return:
        set: of uuids, including associated uuids found AND `updated` uuids
    """
    es = registry[ELASTIC_SEARCH]
    scan_query = {
        'query': {
            'bool': {
                'filter': {
                    'bool': {
                        'should': [
                            {
                                'terms': {
                                    'linked_uuids_embedded.uuid': list(updated)
                                }
                            }
                        ]
                    }
                }
            }
        },
        '_source': {  # in ES7 there is no _type included by default, so inspect embedded.@type
            'includes':
                'item_type'
        }
    }
    if not find_index:
        find_index = get_namespaced_index(registry, '*')
    # size param below == # of results per request, too large and can timeout, too small and will make too
    # many requests - 5000 seems to be a reasonable number. - Will 6/7/21
    results = scan(es, index=find_index, query=scan_query, size=SCAN_PAGE_SIZE)
    invalidated_with_type = {(res['_id'], to_camel_case(res['_source']['item_type'])) for res in results}
    invalidated = {uuid for uuid, _type in invalidated_with_type}

    return (updated | invalidated), invalidated_with_type


def get_uuids_for_types(registry, types=[]):
    """
    WARNING! This makes lots of DB requests and should be used carefully.

    Generator function to return uuids for all the given types. If no
    types provided, uses all types (get all uuids). Because of inheritance
    between item classes, do not iterate over all subtypes (as is done with
    `for uuid in collection`; instead, leverage `collection.iter_no_subtypes`)

    Args:
        registry: the current Registry
        types (list): string item types to specifcally use to find collections.
            Default is empty list, which means all collections are used

    Yields:
        str: uuid of item in collections
    """
    if not isinstance(types, list) or not all(isinstance(t, str) for t in types):  # type check for safety
        raise TypeError('Expected type=list (of strings) for argument "types"')
    collections = registry[COLLECTIONS]
    # might as well sort collections alphabetically, as this was done earlier
    for coll_name in sorted(collections.by_item_type):
        if types and coll_name not in types:
            continue
        for uuid in collections.by_item_type[coll_name].iter_no_subtypes():
            yield str(uuid)


def extract_type_properties(registry, invalidated_item_type):
    """ Helper function, useful for mocking. """
    return registry['types'][invalidated_item_type].schema['properties']


def extract_type_embedded_list(registry, invalidated_item_type):
    """ Helper function, useful for mocking """
    return registry['types'][invalidated_item_type].embedded_list


def extract_type_default_diff(registry, invalidated_item_type):
    """ Helper function that extracts the default diff for this item, if one exists. """
    return getattr(registry['types'][invalidated_item_type], 'default_diff', [])


def extract_base_types(registry, item_type):
    """ Helper function, useful for mocking """
    return registry[TYPES][item_type].base_types


def determine_parent_types(registry, item_type):
    """ Determines the parent types of the given item_type """
    base_types = []
    try:
        base_types = extract_base_types(registry, item_type)
    except KeyError:  # indicative of an error if not testing
        log.info(f'Tried to determine parent type of invalid type: {item_type}')
    return [b for b in base_types if b != 'Item']


def determine_child_types(registry, parent_type):
    """ Determines the child types of the given parent type (to a depth of one). """
    child_types = []
    for potential_child_type, details in registry[TYPES].by_item_type.items():
        if parent_type in getattr(details, 'base_types', []):
            child_types.append(details.name)
    return child_types


def build_diff_metadata(registry, diff):
    """ Helper function for below that builds metadata from diff needed to filter
        invalidation scope.

    :param registry: application registry, used to retrieve type information
    :param diff: a diff of the change (from SQS), see build_diff_from_request
    :returns: 4-tuple:
                * skip bool (to invalidate everything),
                * diff intermediary
                * type that the diff modifies
                * child -> parent type mappings (if they exist, in case we are modifying a leaf type)
    """
    # build representation of diffs
    # item type -> modified fields mapping
    diffs, child_to_parent_type = {}, {}
    skip = False  # if a modified field is a default embed, EVERYTHING has to be invalidated
    modified_item_type = None  # this will get set to the last value in the diff, if any
    for _d in diff:
        modified_item_type, modified_field = _d.split('.', 1)
        if ('.' + modified_field) in DEFAULT_EMBEDS:
            skip = True
            break
        if modified_item_type not in diffs:
            diffs[modified_item_type] = [modified_field]
        else:
            diffs[modified_item_type].append(modified_field)

        default_diff = extract_type_default_diff(registry, modified_item_type)
        if default_diff:
            diffs[modified_item_type].extend(default_diff)

        modified_item_parent_types = determine_parent_types(registry, modified_item_type)
        if modified_item_parent_types:
            child_to_parent_type[modified_item_type] = modified_item_parent_types

    return skip, diffs, modified_item_type, child_to_parent_type


def filter_invalidation_scope(registry, diff, invalidated_with_type, secondary_uuids):
    """ Function that given a diff in the following format:
            ItemType.base_field.terminal_field --> {ItemType: base_field.terminal_field} intermediary
        And a list of invalidated uuids with their type information as a 2-tuple:
            [(<uuid>, <item_type>), ...]
        Removes uuids of item types that were not invalidated from the set of secondary_uuids.

        NOTE: The core logic of invalidation scope occurs in this function.
        The algorithm proceeds roughly as follows:
            for invalidated_item_type:
                for each embed in the invalidated_item_type's embedded list:
                    if this embed does not form a link to the diff target:
                        keep searching
                    else (it forms a candidate link):
                        determine all possible diffs for the valid types
                        if any part of the diff is embedded directly through a terminal field or via *:
                            mark this type as invalidated, cache type result to avoid re-computation
                        else:
                            keep searching embeds
                if we reach this point and have not invalidated, the type is cleared

    :param registry: application registry, used to retrieve type information
    :param diff: a diff of the change (from SQS), see build_diff_from_request
    :param invalidated_with_type: list of 2-tuple (uuid, item_type)
    :param secondary_uuids: primary set of uuids to be invalidated
    :returns: dictionary mapping types to a boolean on whether or not the type is invalidated
    """
    skip, diffs, diff_type, child_to_parent_type = build_diff_metadata(registry, diff)
    valid_diff_types = child_to_parent_type.get(diff_type, []) + [diff_type]
    # go through all invalidated uuids, looking at the embedded list of the item type
    item_type_is_invalidated = {}
    for invalidated_uuid, invalidated_item_type in invalidated_with_type:
        if skip is True:  # if we detected a change to a default embed, invalidate everything

            if invalidated_item_type not in item_type_is_invalidated:
                item_type_is_invalidated[invalidated_item_type] = True
            continue

        # remove this uuid if its item type has been seen before and found to
        # not be invalidated
        if invalidated_item_type in item_type_is_invalidated:
            if item_type_is_invalidated[invalidated_item_type] is False:
                secondary_uuids.discard(invalidated_uuid)
            continue  # nothing else to do here

        # if we get here, we are looking at an invalidated_item_type that exists in the
        # diff, and we need to inspect the embedded list to see if the diff fields are
        # embedded
        properties = extract_type_properties(registry, invalidated_item_type)
        embedded_list = extract_type_embedded_list(registry, invalidated_item_type)

        for embed in embedded_list:

            # check the field up to the embed as this is the path to the linkTo
            # we must determine its type and determine if the given diff could've
            # resulted in an invalidation
            split_embed = embed.split('.')
            link_depth, base_field_item_type = 0, None

            # Checks that schema_part is a linkTo and it is of type item_type
            def is_matched_linkto(schema_part, item_types):
                if 'linkTo' in schema_part and schema_part['linkTo'] in item_types:
                    return True
                return False

            # crawl schema by each increasingly shorter embed searching for where
            # the last link target ends and the terminal field begins
            for i in range(len(split_embed), 0, -1):
                embed_path = '.'.join(split_embed[0:i])
                if embed_path.endswith('*'):
                    continue
                embed_part_schema = crawl_schema(registry['types'], embed_path, properties)
                if is_matched_linkto(embed_part_schema, valid_diff_types):
                    base_field_item_type = embed_part_schema.get('linkTo', None)
                    link_depth = i
                    break
                elif 'items' in embed_part_schema:
                    array = embed_part_schema['items']
                    if is_matched_linkto(array, valid_diff_types):
                        base_field_item_type = array.get('linkTo')
                        link_depth = i
                        break

            # if we did not find a linkTo, this diff is not represented in the current embed
            # so continue looking through the rest of the embeds
            if base_field_item_type is None:
                continue

            # grab terminal field based on the actual linkTo depth
            terminal_field = '.'.join(split_embed[link_depth:])

            # Collect diffs from all possible item_types
            all_possible_diffs = diffs.get(base_field_item_type, [])

            # A linkTo target could be a child type (in that we need to look at parent type diffs as well)
            parent_types = child_to_parent_type.get(base_field_item_type, None)
            if parent_types is not None:
                for parent_type in child_to_parent_type.get(base_field_item_type, []):
                    all_possible_diffs.extend(diffs.get(parent_type, []))

            # It could also be parent type (in that we must look at all potential child types)
            child_types = determine_child_types(registry, base_field_item_type)
            if child_types is not None:
                for child_type in determine_child_types(registry, base_field_item_type) or []:
                    all_possible_diffs.extend(diffs.get(child_type, []))

            if not all_possible_diffs:  # no diffs match this embed
                continue

            # Checks that the 'field' passed is actually contained in the embed list
            def field_is_part_of_target_embed(field):
                terminal_field_as_list = terminal_field.split('.')
                for field_part in field.split('.'):
                    if field_part in terminal_field_as_list:
                        return True
                return False

            # VERY IMPORTANT: for this to work correctly, the fields used in calculated properties MUST
            # be embedded! In addition, if you embed * on a linkTo, modifications to that linkTo will ALWAYS
            # invalidate the item_type
            for field in all_possible_diffs:
                # if terminal field matches a diff exactly, invalidate
                if terminal_field == field:
                    log.info(f'Invalidating item type {invalidated_item_type} based on edit to field {field} given exact'
                             f'embed {terminal_field}')
                # if terminal field is *, invalidate
                elif terminal_field == '*' or (terminal_field.endswith('*') and field_is_part_of_target_embed(field)):
                    log.info(f'Invalidating item type {invalidated_item_type} for field {field} based on star embed {split_embed}')
                # if we edited a link itself or any field on the path, invalidate
                elif field in split_embed:
                    log.info(f'Invalidating item type {invalidated_item_type} based on edit to field {field} given embed'
                             f' path {split_embed}')
                else:
                    log.info(f'Skipping field {field} as {split_embed} does not match')
                    continue
                item_type_is_invalidated[invalidated_item_type] = True
                break

            if item_type_is_invalidated.get(invalidated_item_type):
                break  # if found we don't need to continue searching

        # if we didn't break out of the above loop, we never found an embedded field that was
        # touched, so set this item type to False so all items of this type are NOT invalidated
        if invalidated_item_type not in item_type_is_invalidated:
            secondary_uuids.discard(invalidated_uuid)
            item_type_is_invalidated[invalidated_item_type] = False

    # XXX: Enable to get debugging information on invalidation scope
    # def _sort(tp):
    #     return tp[0]
    # log.error('Diff: %s Invalidated: %s Cleared: %s' % (diffs, sorted(list((k, v) for k, v in item_type_is_invalidated.items()
    #                                                                        if v is True), key=_sort),
    #                                                            sorted(list((k, v) for k, v in item_type_is_invalidated.items()
    #                                                                        if v is False), key=_sort)))
    return item_type_is_invalidated


def _compute_invalidation_scope_base(request, result, source_type, target_type, simulated_prop):
    """ Helper for below route - implements the base case of the API
        Builds a dummy diff from on the simulated prop and determines whether the edit results
        in invalidation of the target type.
    """

    dummy_diff = ['.'.join([source_type, simulated_prop])]
    invalidated_with_type = [('dummy', target_type)]
    invalidated_metadata = filter_invalidation_scope(request.registry, dummy_diff, invalidated_with_type, set())
    if invalidated_metadata.get(target_type, False):
        result['Invalidated'].append(simulated_prop)
    else:
        result['Cleared'].append(simulated_prop)


def _compute_invalidation_scope_recursive(request, result, meta, source_type, target_type, simulated_prop):
    """ Helper for below route - implements the recursive step of the API.
        Traverses the properties computing invalidation scope for all possible patch paths.
    """
    if 'calculatedProperty' in meta:  # we cannot patch calc props, so behavior here is irrelevant
        return
    elif meta['type'] == 'object':
        if 'properties' not in meta:
            return  # sometimes can occur (see workflow.json in fourfront) - nothing we can do
        for sub_prop, sub_meta in meta['properties'].items():
            _compute_invalidation_scope_recursive(request, result, sub_meta, source_type, target_type,
                                                  '.'.join([simulated_prop, sub_prop]))
    elif meta['type'] == 'array':
        sub_type = meta['items']['type']
        if sub_type == 'object':
            if 'properties' not in meta['items']:
                return  # sometimes can occur (see workflow.json in fourfront) - nothing we can do
            for sub_prop, sub_meta in meta['items']['properties'].items():
                _compute_invalidation_scope_recursive(request, result, sub_meta, source_type, target_type,
                                                      '.'.join([simulated_prop, sub_prop]))
        else:
            _compute_invalidation_scope_base(request, result, source_type, target_type, simulated_prop)
    else:
        _compute_invalidation_scope_base(request, result, source_type, target_type, simulated_prop)


@view_config(route_name='compute_invalidation_scope', request_method='POST', permission='index')
@debug_log
def compute_invalidation_scope(context, request):
    """ Computes invalidation scope for a given source item type against a target item type.
        Arguments:
            source_type: item type whose edits we'd like to investigate
            target_type: "impacted" type ie: assume this type was invalidated
        Response:
            source/target type are given back
            Invalidated: list of fields on source_type that, if modified, trigger invalidation of target_type
            Cleared: list of fields on source_type that, if modified, do not trigger invalidation of target_type
    """
    source_type = request.json.get('source_type', None)
    target_type = request.json.get('target_type', None)
    # None-check
    if not source_type or not target_type:
        raise HTTPBadRequest('Missing required parameters: source_type, target_type')
    # Invalid Type
    if source_type not in request.registry[TYPES] or target_type not in request.registry[TYPES]:
        raise HTTPBadRequest('Invalid source/target type: %s/%s' % (source_type, target_type))
    # Abstract type
    # Note 'type' is desired here because concrete types have literal type TypeInfo
    # vs. abstract types have literal type AbstractTypeInfo
    # isinstance() will return True (wrong) since TypeInfo inherits from AbstractTypeInfo
    if type(request.registry[TYPES][source_type]) == AbstractTypeInfo or \
            type(request.registry[TYPES][target_type]) == AbstractTypeInfo:
        raise HTTPBadRequest('One or more of your types is abstract! %s/%s' % (source_type, target_type))
    source_type_schema = request.registry[TYPES][source_type].schema
    result = {
        'Source Type': source_type,
        'Target Type': target_type,
        'Invalidated': [],
        'Cleared': []
    }

    # Walk schema, simulating an edit and computing invalidation scope per field, recording result
    for prop, meta in source_type_schema['properties'].items():
        _compute_invalidation_scope_recursive(request, result, meta, source_type, target_type, prop)

    return result



# -*- coding: utf-8 -*-
"""Load collections and determine the order."""

import gzip
import json
import magic
import mimetypes
import os
import structlog
import webtest
import traceback
import uuid

from base64 import b64encode
from dcicutils.misc_utils import ignored, environ_bool
from dcicutils.secrets_utils import assume_identity
from PIL import Image
from pyramid.paster import get_app
from pyramid.response import Response
from pyramid.view import view_config
from snovault.util import debug_log

from .project import project_filename
from .server_defaults import add_last_modified


text = type(u'')
logger = structlog.getLogger(__name__)


def includeme(config):
    # provide an endpoint to do bulk uploading that just uses loadxl
    config.add_route('load_data', '/load_data')
    config.scan(__name__)


# order of items references with linkTo in a field in  'required' in schemas
# This should be set by the downstream application
ORDER = []

IS_ATTACHMENT = [
    'attachment',
    'file_format_specification',
]


# This uuid should be constant across all portals
LOADXL_USER_UUID = "3202fd57-44d2-44fb-a131-afb1e43d8ae5"


class LoadGenWrapper(object):
    """
    Simple class that accepts a generator function and handles errors by
    setting self.caught to the error message.
    """
    def __init__(self, gen):
        self.gen = gen
        self.caught = None

    def __iter__(self):
        """
        Iterate through self.gen and see if 'ERROR: ' bytes are in any yielded
        value. If so, store the error message as self.caught and raise
        StopIteration to halt the generator.
        """
        # self.caught = yield from self.gen
        for iter_val in self.gen:
            if b'ERROR:' in iter_val:
                self.caught = iter_val.decode()
            yield iter_val

    def close(self):
        if self.caught:
            logger.error('load_data: failed to load with iter_response', error=self.caught)


@view_config(route_name='load_data', request_method='POST', permission='add')
@debug_log
def load_data_view(context, request):
    """
    expected input data

    {'local_path': path to a directory or file in file system
     'fdn_dir': inserts folder under encoded
     'store': if not local_path or fdn_dir, look for a dictionary of items here
     'overwrite' (Bool): overwrite if existing data
     'itype': (list or str): only pick some types from the source or specify type in in_file
     'iter_response': invoke the Response as an app_iter, directly calling load_all_gen
     'config_uri': user supplied configuration file}

    post can contain 2 different styles of data
    1) reference to a folder or file (local_path or fd_dir). If this is done
       itype can be optionally used to specify type of items loaded from files
    2) store in form of {'item_type': [items], 'item_type2': [items]}
       item_type should be same as insert file names i.e. file_fastq
    """
    ignored(context)
    # this is a bit wierd but want to reuse load_data functionality so I'm rolling with it
    config_uri = request.json.get('config_uri', 'production.ini')
    patch_only = request.json.get('patch_only', False)
    post_only = request.json.get('post_only', False)
    app = get_app(config_uri, 'app')
    environ = {'HTTP_ACCEPT': 'application/json', 'REMOTE_USER': 'TEST'}
    testapp = webtest.TestApp(app, environ)
    # expected response
    request.response.status = 200
    result = {
        'status': 'success',
        '@type': ['result'],
    }
    store = request.json.get('store', {})
    local_path = request.json.get('local_path')
    fdn_dir = request.json.get('fdn_dir')
    overwrite = request.json.get('overwrite', False)
    itype = request.json.get('itype')
    iter_resp = request.json.get('iter_response', False)
    inserts = None
    from_json = False
    if fdn_dir:
        inserts = project_filename('tests/data/' + fdn_dir + '/')
    elif local_path:
        inserts = local_path
    elif store:
        inserts = store
        from_json = True
    # if we want to iterate over the response to keep the connection alive
    # this directly calls load_all_gen, instead of load_all
    if iter_resp:
        return Response(
            content_type='text/plain',
            app_iter=LoadGenWrapper(
                load_all_gen(testapp, inserts, None, overwrite=overwrite, itype=itype,
                             from_json=from_json, patch_only=patch_only, post_only=post_only)
            )
        )
    # otherwise, it is a regular view and we can call load_all as usual
    if inserts:
        res = load_all(testapp, inserts, None, overwrite=overwrite, itype=itype, from_json=from_json)
    else:
        res = 'No uploadable content found!'

    if res:  # None if load_all is successful
        print(LOAD_ERROR_MESSAGE)
        request.response.status = 422
        result['status'] = 'error'
        result['@graph'] = str(res)
    return result


def trim(value):
    """Shorten excessively long fields in error log."""
    if isinstance(value, dict):
        return {k: trim(v) for k, v in value.items()}
    if isinstance(value, list):
        return [trim(v) for v in value]
    if isinstance(value, str) and len(value) > 160:
        return value[:77] + '...' + value[-80:]
    return value


def find_doc(docsdir, filename):
    """tries to find the file, if not returns false."""
    path = None
    if not docsdir:
        return
    for dirpath in docsdir:
        candidate = os.path.join(dirpath, filename)
        if not os.path.exists(candidate):
            continue
        if path is not None:
            msg = 'Duplicate filenames: %s, %s' % (path, candidate)
            raise ValueError(msg)
        path = candidate
    if path is None:
        return
    return path


def attachment(path):
    """Create an attachment upload object from a filename Embeds the attachment as a data url."""
    filename = os.path.basename(path)
    mime_type, encoding = mimetypes.guess_type(path)
    major, minor = mime_type.split('/')
    try:
        detected_type = magic.from_file(path, mime=True).decode('ascii')
    except AttributeError:
        detected_type = magic.from_file(path, mime=True)
    # XXX This validation logic should move server-side.
    if not (detected_type == mime_type or
            detected_type == 'text/plain' and major == 'text'):
        raise ValueError('Wrong extension for %s: %s' % (detected_type, filename))
    with open(path, 'rb') as stream:
        attach = {'download': filename,
                  'type': mime_type,
                  'href': 'data:%s;base64,%s' % (mime_type, b64encode(stream.read()).decode('ascii'))}
        if mime_type in ('application/pdf', "application/zip", 'text/plain',
                         'text/tab-separated-values', 'text/html', 'application/msword', 'application/vnd.ms-excel',
                         'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'):
            # XXX Should use chardet to detect charset for text files here.
            return attach
        if major == 'image' and minor in ('png', 'jpeg', 'gif', 'tiff'):
            # XXX we should just convert our tiffs to pngs
            stream.seek(0, 0)
            im = Image.open(stream)
            im.verify()
            if im.format != minor.upper():
                msg = "Image file format %r does not match extension for %s"
                raise ValueError(msg % (im.format, filename))
            attach['width'], attach['height'] = im.size
            return attach
    raise ValueError("Unknown file type for %s" % filename)


def format_for_attachment(json_data, docsdir):
    for field in IS_ATTACHMENT:
        if field in json_data:
            if isinstance(json_data[field], dict):
                pass
            elif isinstance(json_data[field], str):
                path = find_doc(docsdir, json_data[field])
                if not path:
                    del json_data[field]
                    logger.error('Removing {} form {}, expecting path'.format(field, json_data['uuid']))
                else:
                    json_data[field] = attachment(path)
            else:
                # malformatted attachment
                del json_data[field]
                logger.error('Removing {} form {}, expecting path'.format(field, json_data['uuid']))
    return json_data


LOAD_ERROR_MESSAGE = """#   ██▓     ▒█████   ▄▄▄      ▓█████▄  ██▓ ███▄    █   ▄████
#  ▓██▒    ▒██▒  ██▒▒████▄    ▒██▀ ██▌▓██▒ ██ ▀█   █  ██▒ ▀█▒
#  ▒██░    ▒██░  ██▒▒██  ▀█▄  ░██   █▌▒██▒▓██  ▀█ ██▒▒██░▄▄▄░
#  ▒██░    ▒██   ██░░██▄▄▄▄██ ░▓█▄   ▌░██░▓██▒  ▐▌██▒░▓█  ██▓
#  ░██████▒░ ████▓▒░ ▓█   ▓██▒░▒████▓ ░██░▒██░   ▓██░░▒▓███▀▒
#  ░ ▒░▓  ░░ ▒░▒░▒░  ▒▒   ▓▒█░ ▒▒▓  ▒ ░▓  ░ ▒░   ▒ ▒  ░▒   ▒
#  ░ ░ ▒  ░  ░ ▒ ▒░   ▒   ▒▒ ░ ░ ▒  ▒  ▒ ░░ ░░   ░ ▒░  ░   ░
#    ░ ░   ░ ░ ░ ▒    ░   ▒    ░ ░  ░  ▒ ░   ░   ░ ░ ░ ░   ░
#      ░  ░    ░ ░        ░  ░   ░     ░           ░       ░
#                              ░
#   ██▓ ███▄    █   ██████ ▓█████  ██▀███  ▄▄▄█████▓  ██████
#  ▓██▒ ██ ▀█   █ ▒██    ▒ ▓█   ▀ ▓██ ▒ ██▒▓  ██▒ ▓▒▒██    ▒
#  ▒██▒▓██  ▀█ ██▒░ ▓██▄   ▒███   ▓██ ░▄█ ▒▒ ▓██░ ▒░░ ▓██▄
#  ░██░▓██▒  ▐▌██▒  ▒   ██▒▒▓█  ▄ ▒██▀▀█▄  ░ ▓██▓ ░   ▒   ██▒
#  ░██░▒██░   ▓██░▒██████▒▒░▒████▒░██▓ ▒██▒  ▒██▒ ░ ▒██████▒▒
#  ░▓  ░ ▒░   ▒ ▒ ▒ ▒▓▒ ▒ ░░░ ▒░ ░░ ▒▓ ░▒▓░  ▒ ░░   ▒ ▒▓▒ ▒ ░
#   ▒ ░░ ░░   ░ ▒░░ ░▒  ░ ░ ░ ░  ░  ░▒ ░ ▒░    ░    ░ ░▒  ░ ░
#   ▒ ░   ░   ░ ░ ░  ░  ░     ░     ░░   ░   ░      ░  ░  ░
#   ░           ░       ░     ░  ░   ░                    ░
#
#    █████▒▄▄▄       ██▓ ██▓    ▓█████ ▓█████▄
#  ▓██   ▒▒████▄    ▓██▒▓██▒    ▓█   ▀ ▒██▀ ██▌
#  ▒████ ░▒██  ▀█▄  ▒██▒▒██░    ▒███   ░██   █▌
#  ░▓█▒  ░░██▄▄▄▄██ ░██░▒██░    ▒▓█  ▄ ░▓█▄   ▌
#  ░▒█░    ▓█   ▓██▒░██░░██████▒░▒████▒░▒████▓
#   ▒ ░    ▒▒   ▓▒█░░▓  ░ ▒░▓  ░░░ ▒░ ░ ▒▒▓  ▒
#   ░       ▒   ▒▒ ░ ▒ ░░ ░ ▒  ░ ░ ░  ░ ░ ▒  ▒
#   ░ ░     ░   ▒    ▒ ░  ░ ░      ░    ░ ░  ░
#               ░  ░ ░      ░  ░   ░  ░   ░
#                                       ░                    """


def load_all(testapp, inserts, docsdir, overwrite=True, itype=None, from_json=False, patch_only=False, post_only=False,
             skip_types=None):
    """
    Wrapper function for load_all_gen, which invokes the generator returned
    from that function. Takes all of the same args as load_all_gen, so
    please reference that docstring.

    This function uses LoadGenWrapper, which will catch a returned value from
    the execution of the generator, which is an Exception in the case of
    load_all_gen. Return that Exception if encountered, which is consistent
    with the functionality of load_all_gen.
    """
    gen = LoadGenWrapper(
        load_all_gen(testapp, inserts, docsdir, overwrite, itype, from_json, patch_only, post_only, skip_types)
    )
    # run the generator; don't worry about the output
    for _ in gen:
        pass
    # gen.caught is None for success and an error message on failure
    if gen.caught is None:
        return None
    else:
        return Exception(gen.caught)


LOADXL_ALLOW_NONE = environ_bool("LOADXL_ALLOW_NONE", default=True)


def load_all_gen(testapp, inserts, docsdir, overwrite=True, itype=None, from_json=False,
                 patch_only=False, post_only=False, skip_types=None):
    """
    Generator function that yields bytes information about each item POSTed/PATCHed.
    Is the base functionality of load_all function.

    convert data to store format dictionary (same format expected from from_json=True),
    assume main function is to load reasonable number of inserts from a folder

    Args:
        testapp
        inserts : either a folder, file, or a dictionary in the store format
        docsdir : attachment folder
        overwrite (bool)   : if the database contains the item already, skip or patch
        itype (list or str): limit selection to certain type/types
        from_json (bool)   : if set to true, inserts should be dict instead of folder name
        patch_only (bool)  : if set to true will only do second round patch - no posts
        post_only (bool)   : if set to true posts full item no second round or lookup -
                             use with care - will not work if linkTos to items not in db yet
        skip_types (list)  : if set to a list of item files the process will ignore these files
    Yields:
        Bytes with information on POSTed/PATCHed items

    Returns:
        None if successful, otherwise a bytes error message
    """
    if docsdir is None:
        docsdir = []
    # Collect Items
    store = {}
    if from_json:  # we are directly loading json
        store = inserts
    if not from_json:  # we are loading a file
        use_itype = False
        if os.path.isdir(inserts):  # we've specified a directory
            if not inserts.endswith('/'):
                inserts += '/'
            files = [i for i in os.listdir(inserts) if (i.endswith('.json') or i.endswith('.json.gz'))
                     and (i not in skip_types if skip_types else True)]
        elif os.path.isfile(inserts):  # we've specified a single file
            files = [inserts]
            # use the item type if provided AND not a list
            # otherwise guess from the filename
            use_itype = True if (itype and isinstance(itype, str)) else False
        else:  # cannot get the file
            err_msg = 'Failure loading inserts from %s. Could not find matching file or directory.' % inserts
            # import pdb; pdb.set_trace()
            print(err_msg)
            yield str.encode('ERROR: %s\n' % err_msg)
            return
            # raise StopIteration
        # load from the directory/file
        for a_file in files:
            if use_itype:
                item_type = itype
            else:
                item_type = a_file.split('/')[-1].split(".")[0]
                a_file = inserts + a_file
            store[item_type] = get_json_file_content(a_file)

    # if there is a defined set of items, subtract the rest
    if itype:
        if isinstance(itype, list):
            store = {i: store[i] for i in itype if i in store}
        else:
            store = {itype: store.get(itype, [])}
    # clear empty values
    store = {k: v for k, v in store.items() if v is not None}
    if not store:
        if LOADXL_ALLOW_NONE:
            return
        if from_json:
            err_msg = 'No items found in input "store" json'
        else:
            err_msg = 'No items found in %s' % inserts
        if itype:
            err_msg += ' for item type(s) %s' % itype
        # import pdb; pdb.set_trace()
        print(err_msg)
        yield str.encode('ERROR: %s' % err_msg)
        return
        # raise StopIteration
    # order Items
    all_types = list(store.keys())
    for ref_item in reversed(ORDER):
        if ref_item in all_types:
            all_types.insert(0, all_types.pop(all_types.index(ref_item)))
    # collect schemas
    profiles = testapp.get('/profiles/?frame=raw').json

    # run step1 - if item does not exist, post with minimal metadata (and skip indexing since we will patch
    # in round 2)
    second_round_items = {}
    if not patch_only:
        for a_type in all_types:
            first_fields = []
            if not post_only:
                # this conversion of schema name to object type works for all existing schemas at the moment
                obj_type = "".join([i.title() for i in a_type.split('_')])
                # minimal schema
                schema_info = profiles[obj_type]
                req_fields = schema_info.get('required', [])
                ids = schema_info.get('identifyingProperties', [])
                # some schemas did not include aliases
                if 'aliases' not in ids:
                    ids.append('aliases')
                # file format is required for files, but its usability depends this field
                if a_type in ['file_format', 'experiment_type']:
                    req_fields.append('valid_item_types')
                first_fields = list(set(req_fields + ids))
            skip_existing_items = set()
            posted = 0
            skip_exist = 0
            for an_item in store[a_type]:
                exists = False
                if not post_only:
                    try:
                        # 301 because @id is the existing item path, not uuid
                        testapp.get('/'+an_item['uuid'], status=[200, 301])
                        exists = True
                    except Exception:
                        pass
                # skip the items that exists
                # if overwrite=True, still include them in PATCH round
                if exists:
                    skip_exist += 1
                    if not overwrite:
                        skip_existing_items.add(an_item['uuid'])
                    yield str.encode('SKIP: %s\n' % an_item['uuid'])
                else:
                    if post_only:
                        to_post = an_item
                    else:
                        to_post = {key: value for (key, value) in an_item.items() if key in first_fields}
                    to_post = format_for_attachment(to_post, docsdir)
                    try:
                        res = testapp.post_json('/' + a_type + '?skip_indexing=true', to_post)  # skip indexing in round 1
                        assert res.status_code == 201
                        posted += 1
                        # yield bytes to work with Response.app_iter
                        yield str.encode('POST: %s\n' % res.json['@graph'][0]['uuid'])
                    except Exception as e:
                        print('Posting {} failed. Post body:\n{}\nError Message:{}'
                              ''.format(a_type, str(first_fields), str(e)))
                        # remove newlines from error, since they mess with generator output
                        e_str = str(e).replace('\n', '')
                        # import pdb; pdb.set_trace()
                        yield str.encode('ERROR: %s\n' % e_str)
                        return
                        # raise StopIteration
            if not post_only:
                second_round_items[a_type] = [i for i in store[a_type] if i['uuid'] not in skip_existing_items]
            logger.info('{} 1st: {} items posted, {} items exists.'.format(a_type, posted, skip_exist))
            logger.info('{} 1st: {} items will be patched in second round'
                        .format(a_type, str(len(second_round_items.get(a_type, [])))))
    elif overwrite and not post_only:
        logger.info('Posting round skipped')
        for a_type in all_types:
            second_round_items[a_type] = [i for i in store[a_type]]
            logger.info('{}: {} items will be patched in second round'
                        .format(a_type, str(len(second_round_items.get(a_type, [])))))

    # Round II - patch the rest of the metadata (ensuring to index by not passing the query param)
    rnd = ' 2nd' if not patch_only else ''
    for a_type in all_types:
        patched = 0
        if not second_round_items[a_type]:
            logger.info('{}{}: no items to patch'.format(a_type, rnd))
            continue
        for an_item in second_round_items[a_type]:
            an_item = format_for_attachment(an_item, docsdir)
            try:
                add_last_modified(an_item, userid=LOADXL_USER_UUID)
                res = testapp.patch_json('/'+an_item['uuid'], an_item)
                assert res.status_code == 200
                patched += 1
                # yield bytes to work with Response.app_iter
                yield str.encode('PATCH: %s\n' % an_item['uuid'])
            except Exception as e:
                print('Patching {} failed. Patch body:\n{}\n\nError Message:\n{}'.format(
                      a_type, str(an_item), str(e)))
                print('Full error: %s' % traceback.format_exc())
                e_str = str(e).replace('\n', '')
                # import pdb; pdb.set_trace()
                yield str.encode('ERROR: %s\n' % e_str)
                return
                # raise StopIteration
        logger.info('{}{}: {} items patched .'.format(a_type, rnd, patched))

    # explicit return upon finish
    return None


def get_json_file_content(filename):
    """
    Helper function to obtain objects from (compressed) json files.

    :param filename: str file path
    :returns: object loaded from file
    """
    if filename.endswith(".json"):
        with open(filename) as f:
            result = json.loads(f.read())
    elif filename.endswith(".json.gz"):
        with gzip.open(filename) as f:
            result = json.loads(f.read())
    else:
        raise Exception("Expecting a .json or .json.gz file but found %s." % filename)
    return result


def load_data(app, indir='inserts', docsdir=None, overwrite=False,
              use_master_inserts=True, skip_types=None):
    """
    This function will take the inserts folder as input, and place them to the given environment.
    args:
        app:
        indir (inserts): inserts folder, should be relative to tests/data/
        docsdir (None): folder with attachment documents, relative to tests/data
    """
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = webtest.TestApp(app, environ)
    # load master-inserts by default
    if indir != 'master-inserts' and use_master_inserts:
        master_inserts = project_filename('tests/data/master-inserts/')
        master_res = load_all(testapp, master_inserts, [], skip_types=skip_types)
        if master_res:  # None if successful
            print(LOAD_ERROR_MESSAGE)
            logger.error('load_data: failed to load from %s' % master_inserts, error=master_res)
            return master_res

    if not indir.endswith('/'):
        indir += '/'
    inserts = project_filename('tests/data/' + indir)
    if docsdir is None:
        docsdir = []
    else:
        if not docsdir.endswith('/'):
            docsdir += '/'
        docsdir = [project_filename('tests/data/' + docsdir)]
    res = load_all(testapp, inserts, docsdir, overwrite=overwrite)
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_data: failed to load from %s' % docsdir, error=res)
        return res
    return None  # unnecessary, but makes it more clear that no error was encountered


def load_test_data(app, overwrite=False):
    """
    Load inserts and master-inserts

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, docsdir='documents', indir='inserts',
                     overwrite=overwrite)


def load_local_data(app, overwrite=False):
    """
    Load inserts from temporary insert folders, if present and populated
    with .json insert files.
    If not present, load inserts and master-inserts.

    Returns:
        None if successful, otherwise Exception encountered
    """

    test_insert_dirs = [
        'temp-local-inserts',
        'demo_inserts'
    ]

    for test_insert_dir in test_insert_dirs:
        chk_dir = project_filename("tests/data/" + test_insert_dir)
        for (dirpath, dirnames, filenames) in os.walk(chk_dir):
            if any([fn for fn in filenames if fn.endswith('.json') or fn.endswith('.json.gz')]):
                logger.info('Loading inserts from "{}" directory.'.format(test_insert_dir))
                return load_data(app, docsdir='documents', indir=test_insert_dir, use_master_inserts=True,
                                 overwrite=overwrite)

    # Default to 'inserts' if no temp inserts found.
    return load_data(app, docsdir='documents', indir='inserts', use_master_inserts=True, overwrite=overwrite)


def load_prod_data(app, overwrite=False):
    """
    Load master-inserts

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, indir='master-inserts', overwrite=overwrite)


def load_deploy_data(app, overwrite=True, **kwargs):
    """
    Load deploy-inserts and master-inserts. Overwrites duplicate items
    in both directories to match deploy-inserts version.

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, docsdir='documents', indir="deploy-inserts", overwrite=True)


# Set of emails required by the application to function
REQUIRED_USER_CONFIG = [
    {
        'email': 'loadxl@hms.harvard.edu',
        'first_name': 'loadxl',
        'last_name': 'loadxl',
        'uuid': '3202fd57-44d2-44fb-a131-afb1e43d8ae5'
    },
    {
        'email': 'cgap.platform@gmail.com',
        'first_name': 'Platform',
        'last_name': 'Admin',
        'uuid': 'b5f738b6-455a-42e5-bc1c-77fbfd9b15d2'
    },
    {
        'email': 'foursight.app@gmail.com',
        'first_name': 'Foursight',
        'last_name': 'App',
        'uuid': '7677f8a8-79d2-4cff-ab0a-a967a2a68e39'
    },
    {
        'email': 'tibanna.app@gmail.com',
        'first_name': 'Tibanna',
        'last_name': 'App',
        'uuid': 'b041dba8-e2b2-4e54-a621-97edb508a0c4'
    },
]


def load_custom_data(app, overwrite=False):
    """
    Load deploy-inserts and master-inserts, EXCEPT instead of loading the default user.json,
    generate users (if they do not already exist) from the ENCODED_ADMIN_USERS setting in
    the GAC. We assume it has structure consistent with what the template will build in 4dn-cloud-infra
    ie:
        [{"first_name": "John", "last_name": "Doe", "email": "john_doe@example.com"}]
    """
    # start with the users
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = webtest.TestApp(app, environ)
    identity = assume_identity()
    admin_users = json.loads(identity.get('ENCODED_ADMIN_USERS', '{}'))
    if not admin_users:  # we assume you must have set one of these
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_custom_data: failed to load users as none were set - ensure GAC value'
                     ' ENCODED_ADMIN_USERS is set and formatted correctly!')
        return admin_users

    # post all users
    for user in (admin_users + REQUIRED_USER_CONFIG):
        try:
            first_name, last_name, email, _uuid = (user['first_name'], user['last_name'], user['email'],
                                                   user.get('uuid', str(uuid.uuid4())))
        except KeyError:
            print(LOAD_ERROR_MESSAGE)
            logger.error('load_custom_data: failed to load users as they were malformed - ensure GAC value'
                         ' ENCODED_ADMIN_USERS is set, has type array and consists of objects all containing keys'
                         ' and values for first_name, last_name and email!')
            return user
        item = {
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
            'groups': ['admin'],
            'uuid': _uuid
        }
        testapp.post_json('/User', item, status=201)

    res = load_data(app, docsdir='documents', indir='deploy-inserts', overwrite=overwrite, skip_types=['user.json'])
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_custom_data: failed to load from deploy-inserts', error=res)
        return res

    return None


def load_cypress_data(app, overwrite=False):
    """
    Load master-inserts and cypress-test-inserts.
    By default, does not overwrite duplicate items in both directories

    Returns:
        None if successful, otherwise Exception encountered
    """
    return load_data(app, indir='cypress-test-inserts', overwrite=overwrite)


def load_data_by_type(app, indir='master-inserts', overwrite=True, itype=None):
    """
    This function will load inserts of type itype from the indir directory.
    args:
        indir (inserts): inserts folder, should be relative to tests/data/
        itype: item type to load (e.g. "higlass_view_config")
    """

    if itype is None:
        print('load_data_by_type: No item type specified. Not loading anything.')
        return

    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = webtest.TestApp(app, environ)

    if not indir.endswith('/'):
        indir += '/'
    inserts = project_filename('tests/data/' + indir)

    res = load_all(testapp, inserts, docsdir=[], overwrite=overwrite, itype=itype)
    if res:  # None if successful
        print(LOAD_ERROR_MESSAGE)
        logger.error('load_data_by_type: failed to load from %s' % indir, error=res)
        return res
    return None  # unnecessary, but makes it more clear that no error was encountered



import structlog
import logging
import argparse
import json
from os import walk
# use ff_utils to find inserts and write data
from dcicutils.ff_utils import search_metadata, expand_es_metadata, dump_results_to_json
# use this function to read inserts
from .run_upgrader_on_inserts import get_inserts

logger = structlog.getLogger(__name__)
EPILOG = __doc__


def read_local_inserts_dir(dir_name, path, target_types=[]):
    """
    Given path string path, read local inserts directory and return a
    dictionary of all inserts keyed by item type, as well as a list of all
    found uuids

    Args:
        dir_name (str): string name of the inserts directory
        path (str): string path to the inserts directory
        target_types (list): list of item types to load. Empty means all types

    Returns:
        dict of inserts, list of item uuids
    """
    item_types = []
    item_uuids = []
    local_inserts = {}
    # find item types that are represented in the given inserts path
    for (dirpath, dirnames, filenames) in walk(path):
        item_types = [it[:-5] for it in filenames if it.endswith('.json')]
    if target_types:
        bad_item_types = [it for it in target_types if it not in item_types]
        if bad_item_types:
            raise Exception('update_inserts: Specified item type(s) %s are not found in '
                            'the inserts dir. Found: %s' % (bad_item_types, item_types))
    # update item_types if user specified specific ones
    fetch_item_types = target_types if target_types else item_types
    # load current insert contents from json file
    for item_type in item_types:
        local_inserts[item_type] = {}  # key these by uuid for now
        for it_item in get_inserts(dir_name, item_type):
            # only fetch items for specified fetch_item_types
            if item_type in fetch_item_types:
                item_uuids.append(it_item['uuid'])
            local_inserts[item_type][it_item['uuid']] = it_item
    return local_inserts, item_uuids


def main():
    """
    Use this command to update the inserts from a given fourfront env
    """
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Update Inserts", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--env', default='data',
                        help='FF environment to update from. Defaults to data')
    parser.add_argument('--dest', default='temp-local-inserts',
                        help="destination file in inserts dir to write to")
    parser.add_argument('--item-type', action='append', default=[],
                        help="item type, e.g. file_fastq. Defaults to all types")
    parser.add_argument('--ignore-field', action='append', default=["submitted_by", "date_created", "last_modified", "schema_version"],
                        help='field name to ignore when running expand_es_metadata')
    parser.add_argument('--from-search', help='query passed to search_metadata to find uuids')

    args = parser.parse_args()
    # this will work since bin/ commands are run from root FF directory
    inserts_location = 'src/encoded/tests/data'
    # hardcode these to prevent accidental creation of inserts files
    inserts_files = ['inserts', 'master-inserts', 'perf-testing', 'workbook-inserts',
                     'temp-local-inserts', 'deploy-inserts']
    if args.dest not in inserts_files:
        raise Exception('Specified inserts destination %s must be one of: %s'
                        % (args.dest, inserts_files))
    inserts_path = '/'.join([inserts_location, args.dest])

    local_inserts, item_uuids = read_local_inserts_dir(args.dest, inserts_path, args.item_type)

    # Used to preserve order of existing inserts in folder(s), if any.
    local_inserts_ordering_map = {}
    for item_type, local_inserts_for_type in local_inserts.items():
        for insrt_index, insrt_uuid in enumerate(local_inserts_for_type):
            # Duplicate insrt_indx between different item types are OK and present.
            # local_inserts_ordering_map is shallow.
            local_inserts_ordering_map[insrt_uuid] = insrt_index

    # add uuids from the input search result, if present
    if args.from_search:
        use_search = args.from_search
        # get frame=object search results to keep response small
        if 'frame=' not in use_search:
            use_search += '&frame=object'
        search_res = search_metadata(use_search, ff_env=args.env)
        search_uuids = [item['uuid'] for item in search_res]
        logger.info('update_inserts: Will update using %s items from search' % len(search_uuids))
        item_uuids = list(set(item_uuids + search_uuids))

    # now find uuids and all linked from the given server
    svr_inserts, svr_uuids = expand_es_metadata(item_uuids, ff_env=args.env,
                                                store_frame='raw', add_pc_wfr=True,
                                                ignore_field=args.ignore_field)

    # if we are updating `inserts`, must make sure that items don't conflict
    # with those in `master-inserts`
    skip_uuids = set()
    if args.dest == 'inserts':
        master_path = '/'.join([inserts_location, 'master-inserts'])
        master_inserts, master_uuids = read_local_inserts_dir('master-inserts', master_path)
        item_conflict_report = {}
        for item_type in svr_inserts:
            itype_err = []
            itype_okay = []
            conflicting_items = [item for item in svr_inserts[item_type] if item['uuid'] in master_uuids]
            for conflict in conflicting_items:
                # compare inserts by loading json objects
                svr_json = json.dumps(conflict, sort_keys=True)
                mstr_json = json.dumps(master_inserts[item_type][conflict['uuid']], sort_keys=True)
                if svr_json != mstr_json:
                    itype_err.append(conflict['uuid'])
                else:
                    # the json is the same. Remove from the `inserts` update
                    skip_uuids.add(conflict['uuid'])
                    itype_okay.append(conflict['uuid'])
            item_conflict_report[item_type] = {'error': itype_err, 'okay': itype_okay}
        if any([it for it in item_conflict_report if item_conflict_report[it]['error']]):
            error_report = {it: item_conflict_report[it]['error'] for it in item_conflict_report}
            logger.error('update_inserts: Cannot update the following items in "inserts" directory,'
                            ' since there are conflicting items with different values'
                            'in the master-inserts. Update those first. Conflicts:\n%s' % json.dumps(error_report, indent=4))
            raise Exception('Cannot load inserts as there are conflicting items in `master-inserts`')
        elif any([it for it in item_conflict_report if item_conflict_report[it]['okay']]):
            conflict_report = {it: item_conflict_report[it]['okay'] for it in item_conflict_report}
            logger.warning('update_inserts: The following items are already in "master-inserts".'
                           ' Will not add to "inserts". Items:\n%s' % json.dumps(conflict_report, indent=4))

    # now we need to update the server inserts with contents from local inserts
    # so that existing information is not lost
    for item_type in svr_inserts:
        if skip_uuids:
            # remove items specified by skip uuids
            svr_inserts[item_type] = [
                insrt for insrt in svr_inserts[item_type]
                if insrt['uuid'] not in skip_uuids
            ]
        svr_inserts[item_type].sort(key=lambda insrt: local_inserts_ordering_map.get(insrt["uuid"], 99999) )
        for item_uuid in local_inserts.get(item_type, {}):
            if item_uuid not in svr_uuids and item_uuid not in skip_uuids:
                svr_inserts[item_type].append(local_inserts[item_type][item_uuid])

    dump_results_to_json(svr_inserts, inserts_path)
    logger.info('update_inserts: Successfully wrote to %s' % inserts_path)
    for item_type in svr_inserts:
        logger.info('update_inserts: Wrote %s items to %s' %
                    (len(svr_inserts[item_type]), item_type + '.json'))


if __name__ == "__main__":
    main()



"""base class creation for all the schemas that exist."""
from pyramid.security import (
    # ALL_PERMISSIONS,
    Allow,
    Deny,
    # DENY_ALL,
    Everyone,
)
from pyramid.view import (
    view_config,
)
from .. import Item, Collection, AbstractCollection, abstract_collection, calculated_property
from ..util import debug_log
from ..validators import (
    validate_item_content_post,
    validate_item_content_put,
    validate_item_content_patch,
    validate_item_content_in_place,
    no_validate_item_content_post,
    no_validate_item_content_put,
    no_validate_item_content_patch
)
from ..crud_views import (
    collection_add as sno_collection_add,
    item_edit as sno_item_edit,
)
from ..interfaces import CONNECTION
from typing import Any, List, Tuple, Union
from ..server_defaults import get_userid, add_last_modified


Acl = List[Tuple[Any, Any, Union[str, List[str]]]]

# Item acls
# TODO (C4-332): consolidate all acls into one place - i.e. their own file
ONLY_ADMIN_VIEW_ACL: Acl = [
    (Allow, 'group.admin', ['view', 'edit']),
    (Allow, 'group.read-only-admin', ['view']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Deny, Everyone, ['view', 'edit'])
]


PUBLIC_ACL: Acl = [
    (Allow, Everyone, ['view']),
] + ONLY_ADMIN_VIEW_ACL


DELETED_ACL: Acl = [
    (Deny, Everyone, 'visible_for_edit')
] + ONLY_ADMIN_VIEW_ACL


# Used for 'draft' status
ALLOW_OWNER_EDIT: Acl = [
    (Allow, 'role.owner', ['view', 'edit']),
] + ONLY_ADMIN_VIEW_ACL


def get_item_or_none(request, value, itype=None, frame='object'):
    """
    Return the view of an item with given frame. Can specify different types
    of `value` for item lookup

    Args:
        request: the current Request
        value (str): String item identifier or a dict containing @id/uuid
        itype (str): Optional string collection name for the item (e.g. /file-formats/)
        frame (str): Optional frame to return. Defaults to 'object'

    Returns:
        dict: given view of the item or None on failure
    """
    item = None

    if isinstance(value, dict):
        if 'uuid' in value:
            value = value['uuid']
        elif '@id' in value:
            value = value['@id']

    svalue = str(value)

    # Below case is for UUIDs & unique_keys such as accessions, but not @ids
    if not svalue.startswith('/') and not svalue.endswith('/'):
        svalue = '/' + svalue + '/'
        if itype is not None:
            svalue = '/' + itype + svalue

    # Request.embed will attempt to get from ES for frame=object/embedded
    # If that fails, get from DB. Use '@@' syntax instead of 'frame=' because
    # these paths are cached in indexing
    try:
        item = request.embed(svalue, '@@' + frame)
    except Exception:
        pass

    # could lead to unexpected errors if == None
    return item


def validate_item_type_of_linkto_field(context, request):
    """We are doing this case by case on item specific types files,
    but might want to carry it here if filter is used more often.
    If any of the submitted fields contain an ff_flag property starting with "filter",
    the field in the filter is used for validating the type of the linked item.
    Example: file has field file_format which is a linkTo FileFormat.
    FileFormat items contain a field called "valid_item_types".
    We have the ff_flag on file_format field called "filter:valid_item_types"."""
    pass


# ----------
# Common lists of embeds to be re-used in certain files (similar to schema mixins)
# ----------

static_content_embed_list = [
    "static_headers.*",            # Type: UserContent, may have differing properties
    "static_content.content.@type",
    "static_content.content.content",
    "static_content.content.name",
    "static_content.content.title",
    "static_content.content.status",
    "static_content.content.description",
    "static_content.content.options",
    "static_content.content.institution",
    "static_content.content.project",
    "static_content.content.filetype"
]


class AbstractCollection(AbstractCollection):
    """smth."""

    def __init__(self, *args, **kw):
        try:
            self.lookup_key = kw.pop('lookup_key')
        except KeyError:
            pass
        super(AbstractCollection, self).__init__(*args, **kw)

    def get(self, name, default=None):
        """
        heres' and example of why this is the way it is:
        ontology terms have uuid or term_id as unique ID keys
        and if neither of those are included in post, try to
        use term_name such that:
        No - fail load with non-existing term message
        Multiple - fail load with ‘ambiguous name - more than 1 term with that name exist use ID’
        Single result - get uuid and use that for post/patch
        """
        resource = super(AbstractCollection, self).get(name, None)
        if resource is not None:
            return resource
        if ':' in name:
            resource = self.connection.get_by_unique_key('alias', name)
            if resource is not None:
                if not self._allow_contained(resource):
                    return default
                return resource
        if getattr(self, 'lookup_key', None) is not None:
            # lookup key translates to query json by key / value and return if only one of the
            # item type was found... so for keys that are mostly unique, but do to whatever
            # reason (bad data mainly..) can be defined as unique keys
            item_type = self.type_info.item_type
            resource = self.connection.get_by_json(self.lookup_key, name, item_type)
            if resource is not None:
                if not self._allow_contained(resource):
                    return default
                return resource
        return default


class Collection(Collection, AbstractCollection):
    """smth."""

    def __init__(self, *args, **kw):
        """smth."""
        super(Collection, self).__init__(*args, **kw)
        if hasattr(self, '__acl__'):
            return


@abstract_collection(
    name='items',
    properties={
        'title': "Item Listing",
        'description': 'Abstract collection of all Items.',
    })
class Item(Item):
    """smth."""
    item_type = 'item'
    AbstractCollection = AbstractCollection
    Collection = Collection
    STATUS_ACL = {  # note that this should ALWAYS be overridden by downstream application
        # standard_status
        'shared': PUBLIC_ACL,
        'obsolete': PUBLIC_ACL,
        'current': PUBLIC_ACL,
        'inactive': PUBLIC_ACL,
        'in review': PUBLIC_ACL,
        'uploaded': PUBLIC_ACL,
        'uploading': PUBLIC_ACL,
        'archived': PUBLIC_ACL,
        'deleted': DELETED_ACL,
        'replaced': ONLY_ADMIN_VIEW_ACL,
        'public': PUBLIC_ACL,
        'draft': ALLOW_OWNER_EDIT
    }
    FACET_ORDER_OVERRIDE = {}  # empty by default

    # Items of these statuses are filtered out from rev links
    filtered_rev_statuses = ('deleted')

    # Default embed list for all encoded Items
    embedded_list = static_content_embed_list

    def __init__(self, registry, models):
        super().__init__(registry, models)
        self.STATUS_ACL = self.__class__.STATUS_ACL

    @property
    def __name__(self):
        """smth."""
        if self.name_key is None:
            return self.uuid
        properties = self.upgrade_properties()
        if properties.get('status') == 'replaced':
            return self.uuid
        return properties.get(self.name_key, None) or self.uuid

    def __acl__(self):
        """This sets the ACL for the item based on mapping of status to ACL.
           If there is no status or the status is not included in the STATUS_ACL
           lookup then the access is set to admin only
        """
        # Don't finalize to avoid validation here.
        properties = self.upgrade_properties().copy()
        status = properties.get('status')
        return self.STATUS_ACL.get(status, ONLY_ADMIN_VIEW_ACL)

    def __ac_local_roles__(self):
        """Adds additional information allowing access of the Item based on
           properties of the Item - currently most important is Project.
           eg. ITEM.__ac_local_roles = {
                    institution.uuid: role.institution_member,
                    project.uuid: role.project_member
                }
          """
        roles = {}
        properties = self.upgrade_properties()
        if 'institution' in properties:
            # add institution_member as well
            inst_member = 'institution.%s' % properties['institution']
            roles[inst_member] = 'role.institution_member'
            # to avoid conflation of the project used for attribution of the User ITEM
            # from the project(s) specified in the project_roles specifying project_editor
            # role - instead of using 'bare' project
        if 'project' in properties:
            project_editors = 'editor_for.%s' % properties['project']
            roles[project_editors] = 'role.project_editor'
        # This emulates __ac_local_roles__ of User.py (role.owner) - taken from 4DN in 2022-01
        if 'submitted_by' in properties:
            submitter = 'userid.%s' % properties['submitted_by']
            roles[submitter] = 'role.owner'
        return roles

    def add_accession_to_title(self, title):
        if self.properties.get('accession') is not None:
            return title + ' - ' + self.properties.get('accession')
        return title

    def unique_keys(self, properties):
        """smth."""
        keys = super(Item, self).unique_keys(properties)
        if 'accession' not in self.schema['properties']:
            return keys
        keys.setdefault('accession', []).extend(properties.get('alternate_accessions', []))
        if properties.get('status') != 'replaced' and 'accession' in properties:
            keys['accession'].append(properties['accession'])
        return keys

    def is_update_by_admin_user(self):
        # determine if the submitter in the properties is an admin user
        userid = get_userid()
        users = self.registry['collections']['User']
        user = users.get(userid)
        if 'groups' in user.properties:
            if 'admin' in user.properties['groups']:
                return True
        return False

    def _update(self, properties, sheets=None):
        add_last_modified(properties)
        super(Item, self)._update(properties, sheets)

    @calculated_property(schema={
        "title": "Display Title",
        "description": "A calculated title for every object in 4DN",
        "type": "string"
    })
    def display_title(self, request=None):
        """create a display_title field."""
        display_title = ""
        look_for = [
            "title",
            "name",
            "location_description",
            "accession",
        ]
        properties = self.upgrade_properties()
        for field in look_for:
            # special case for user: concatenate first and last names
            display_title = properties.get(field, None)
            if display_title:
                if field != 'accession':
                    display_title = self.add_accession_to_title(display_title)
                return display_title
        # if none of the existing terms are available, use @type + date_created
        try:
            type_date = self.__class__.__name__ + " from " + properties.get("date_created", None)[:10]
            return type_date
        # last resort, use uuid
        except Exception:
            return properties.get('uuid', None)

    def rev_link_atids(self, request, rev_name):
        """
        Returns the list of reverse linked items given a defined reverse link,
        which should be formatted like:
        rev = {
            '<reverse field name>': ('<reverse item class>', '<reverse field to find>'),
        }

        """
        conn = request.registry[CONNECTION]
        return [request.resource_path(conn[uuid]) for uuid in
                self.get_filtered_rev_links(request, rev_name)]


@calculated_property(context=Item.AbstractCollection, category='action')
def add(context, request):
    """smth."""
    if request.has_permission('add', context):
        type_name = context.type_info.name
        return {
            'name': 'add',
            'title': 'Add',
            'profile': '/profiles/{name}.json'.format(name=type_name),
            'href': '/search/?type={name}&currentAction=add'.format(name=type_name),
        }


@calculated_property(context=Item, category='action')
def edit(context, request):
    """smth."""
    if request.has_permission('edit'):
        return {
            'name': 'edit',
            'title': 'Edit',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}?currentAction=edit'.format(item_uri=request.resource_path(context)),
        }


@calculated_property(context=Item, category='action')
def create(context, request):
    if request.has_permission('create'):
        return {
            'name': 'create',
            'title': 'Create',
            'profile': '/profiles/{ti.name}.json'.format(ti=context.type_info),
            'href': '{item_uri}?currentAction=create'.format(item_uri=request.resource_path(context)),
        }


@view_config(
    context=Collection,
    permission='add',
    request_method='POST',
    # validators=[]  # TURNS OFF VALIDATION HERE ([validate_item_content_post] previously)
    validators=[validate_item_content_post]
)
@view_config(
    context=Collection,
    permission='add_unvalidated',
    request_method='POST',
    validators=[no_validate_item_content_post],
    request_param=['validate=false']
)
@debug_log
def collection_add(context, request, render=None):

    # institution_needed = False
    # project_needed = False
    # data = request.json
    # schema = context.type_info.schema
    #
    # required_properties = schema.get("required", [])
    # if "institution" in required_properties and "institution" not in data:
    #     institution_needed = True
    #
    # if "project" in required_properties and "project" not in data:
    #     project_needed = True
    #
    # if request.authenticated_userid and (institution_needed or project_needed):
    #     namespace, userid = request.authenticated_userid.split(".", 1)
    #     user_item = get_item_or_none(request, userid, itype="/users/", frame="object")
    #     new_data = data.copy()
    #     if institution_needed and "institution" in user_item:
    #         new_data["institution"] = user_item["institution"]
    #     if project_needed and "project" in user_item:
    #         new_data["project"] = user_item["project"]
    #
    #     # Override initial JSON body of request (hacky? better way?)
    #     setattr(request, "json", new_data)
    #
    # # Perform validation that would occur otherwise
    # validate_item_content_post(context, request)
    # if request.errors:
    #     return HTTPUnprocessableEntity(
    #         json={'errors': request.errors},
    #         content_type='application/json'
    #     )
    return sno_collection_add(context, request, render)


@view_config(context=Item, permission='edit', request_method='PUT',
             validators=[validate_item_content_put])
@view_config(context=Item, permission='edit', request_method='PATCH',
             validators=[validate_item_content_patch])
@view_config(context=Item, permission='edit_unvalidated', request_method='PUT',
             validators=[no_validate_item_content_put],
             request_param=['validate=false'])
@view_config(context=Item, permission='edit_unvalidated', request_method='PATCH',
             validators=[no_validate_item_content_patch],
             request_param=['validate=false'])
@view_config(context=Item, permission='index', request_method='GET',
             validators=[validate_item_content_in_place],
             request_param=['check_only=true'])
@debug_log
def item_edit(context, request, render=None):
    # This works
    # Probably don't need to extend re: institution + project since if editing, assuming these have previously existed.
    return sno_item_edit(context, request, render)



import humanfriendly
import logging
import psutil


# https://code.google.com/p/modwsgi/wiki/RegisteringCleanupCode


class Generator2:
    def __init__(self, iterable, callback, environ):
        self.__iterable = iterable
        self.__callback = callback
        self.__environ = environ

    def __iter__(self):
        for item in self.__iterable:
            yield item

    def close(self):
        try:
            if hasattr(self.__iterable, 'close'):
                self.__iterable.close()
        finally:
            self.__callback(self.__environ)


class ExecuteOnCompletion2:
    def __init__(self, application, callback):
        self.__application = application
        self.__callback = callback

    def __call__(self, environ, start_response):
        try:
            result = self.__application(environ, start_response)
        except BaseException:
            self.__callback(environ)
            raise
        return Generator2(result, self.__callback, environ)


def rss_checker(rss_limit=None, rss_percent_limit=None):
    """
    Uses a configured rss_limit (absolute amount in bytes) and percentage
    rss_limit to determine whether to kill the running process.
    If the current rss usage is above rss_limit AND the percentage rss usage
    of physical memory is above rss_percent_limit, kill the process
    """
    log = logging.getLogger(__name__)
    process = psutil.Process()

    def callback(environ):
        rss = process.memory_info().rss
        over_rss = rss_limit and rss > rss_limit
        rss_perc = process.memory_percent(memtype="rss")  # XXX: this does not work on Fargate (reports host stats)
        if rss_percent_limit:
            over_perc = rss_perc > rss_percent_limit
        else:
            over_perc = True  # only consider rss if we have no percent set
        if over_rss and over_perc:
            log.error(f"Killing process. Memory usage: {rss}Mb (limit {rss_limit}); Percentage "
                      f"{rss_perc} (limit {rss_percent_limit})")
            process.kill()

    return callback


def filter_app(app, global_conf, rss_limit=None, rss_percent_limit=None):
    if rss_limit is not None:
        rss_limit = humanfriendly.parse_size(rss_limit)
    if rss_percent_limit is not None:
        rss_percent_limit = float(rss_percent_limit)

    callback = rss_checker(rss_limit, rss_percent_limit)
    return ExecuteOnCompletion2(app, callback)
